# distutils: libraries = flint
# distutils: depends = flint/ca.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void ca_ctx_init(ca_ctx_t ctx) noexcept
    void ca_ctx_clear(ca_ctx_t ctx) noexcept
    void ca_ctx_print(ca_ctx_t ctx) noexcept
    void ca_init(ca_t x, ca_ctx_t ctx) noexcept
    void ca_clear(ca_t x, ca_ctx_t ctx) noexcept
    void ca_swap(ca_t x, ca_t y, ca_ctx_t ctx) noexcept
    void ca_get_fexpr(fexpr_t res, const ca_t x, ulong flags, ca_ctx_t ctx) noexcept
    int ca_set_fexpr(ca_t res, const fexpr_t expr, ca_ctx_t ctx) noexcept
    void ca_print(const ca_t x, ca_ctx_t ctx) noexcept
    void ca_fprint(FILE * fp, const ca_t x, ca_ctx_t ctx) noexcept
    char * ca_get_str(const ca_t x, ca_ctx_t ctx) noexcept
    void ca_printn(const ca_t x, slong n, ca_ctx_t ctx) noexcept
    void ca_zero(ca_t res, ca_ctx_t ctx) noexcept
    void ca_one(ca_t res, ca_ctx_t ctx) noexcept
    void ca_neg_one(ca_t res, ca_ctx_t ctx) noexcept
    void ca_i(ca_t res, ca_ctx_t ctx) noexcept
    void ca_neg_i(ca_t res, ca_ctx_t ctx) noexcept
    void ca_pi(ca_t res, ca_ctx_t ctx) noexcept
    void ca_pi_i(ca_t res, ca_ctx_t ctx) noexcept
    void ca_euler(ca_t res, ca_ctx_t ctx) noexcept
    void ca_unknown(ca_t res, ca_ctx_t ctx) noexcept
    void ca_undefined(ca_t res, ca_ctx_t ctx) noexcept
    void ca_uinf(ca_t res, ca_ctx_t ctx) noexcept
    void ca_pos_inf(ca_t res, ca_ctx_t ctx) noexcept
    void ca_neg_inf(ca_t res, ca_ctx_t ctx) noexcept
    void ca_pos_i_inf(ca_t res, ca_ctx_t ctx) noexcept
    void ca_neg_i_inf(ca_t res, ca_ctx_t ctx) noexcept
    void ca_set(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_set_si(ca_t res, slong v, ca_ctx_t ctx) noexcept
    void ca_set_ui(ca_t res, ulong v, ca_ctx_t ctx) noexcept
    void ca_set_fmpz(ca_t res, const fmpz_t v, ca_ctx_t ctx) noexcept
    void ca_set_fmpq(ca_t res, const fmpq_t v, ca_ctx_t ctx) noexcept
    void ca_set_d(ca_t res, double x, ca_ctx_t ctx) noexcept
    void ca_set_d_d(ca_t res, double x, double y, ca_ctx_t ctx) noexcept
    void ca_transfer(ca_t res, ca_ctx_t res_ctx, const ca_t src, ca_ctx_t src_ctx) noexcept
    void ca_set_qqbar(ca_t res, const qqbar_t x, ca_ctx_t ctx) noexcept
    int ca_get_fmpz(fmpz_t res, const ca_t x, ca_ctx_t ctx) noexcept
    int ca_get_fmpq(fmpq_t res, const ca_t x, ca_ctx_t ctx) noexcept
    int ca_get_qqbar(qqbar_t res, const ca_t x, ca_ctx_t ctx) noexcept
    int ca_can_evaluate_qqbar(const ca_t x, ca_ctx_t ctx) noexcept
    void ca_randtest_rational(ca_t res, flint_rand_t state, slong bits, ca_ctx_t ctx) noexcept
    void ca_randtest(ca_t res, flint_rand_t state, slong depth, slong bits, ca_ctx_t ctx) noexcept
    void ca_randtest_special(ca_t res, flint_rand_t state, slong depth, slong bits, ca_ctx_t ctx) noexcept
    void ca_randtest_same_nf(ca_t res, flint_rand_t state, const ca_t x, slong bits, slong den_bits, ca_ctx_t ctx) noexcept
    bint ca_equal_repr(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    int ca_cmp_repr(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    ulong ca_hash_repr(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_unknown(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_special(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_qq_elem(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_qq_elem_zero(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_qq_elem_one(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_qq_elem_integer(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_nf_elem(const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_cyclotomic_nf_elem(slong * p, ulong * q, const ca_t x, ca_ctx_t ctx) noexcept
    bint ca_is_generic_elem(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_number(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_zero(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_one(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_neg_one(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_i(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_neg_i(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_algebraic(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_rational(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_integer(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_real(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_negative_real(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_imaginary(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_undefined(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_infinity(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_uinf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_signed_inf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_pos_inf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_neg_inf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_pos_i_inf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_is_neg_i_inf(const ca_t x, ca_ctx_t ctx) noexcept
    truth_t ca_check_equal(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    truth_t ca_check_lt(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    truth_t ca_check_le(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    truth_t ca_check_gt(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    truth_t ca_check_ge(const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_merge_fields(ca_t resx, ca_t resy, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_condense_field(ca_t res, ca_ctx_t ctx) noexcept
    ca_ext_ptr ca_is_gen_as_ext(const ca_t x, ca_ctx_t ctx) noexcept
    void ca_neg(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_add_fmpq(ca_t res, const ca_t x, const fmpq_t y, ca_ctx_t ctx) noexcept
    void ca_add_fmpz(ca_t res, const ca_t x, const fmpz_t y, ca_ctx_t ctx) noexcept
    void ca_add_ui(ca_t res, const ca_t x, ulong y, ca_ctx_t ctx) noexcept
    void ca_add_si(ca_t res, const ca_t x, slong y, ca_ctx_t ctx) noexcept
    void ca_add(ca_t res, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_sub_fmpq(ca_t res, const ca_t x, const fmpq_t y, ca_ctx_t ctx) noexcept
    void ca_sub_fmpz(ca_t res, const ca_t x, const fmpz_t y, ca_ctx_t ctx) noexcept
    void ca_sub_ui(ca_t res, const ca_t x, ulong y, ca_ctx_t ctx) noexcept
    void ca_sub_si(ca_t res, const ca_t x, slong y, ca_ctx_t ctx) noexcept
    void ca_fmpq_sub(ca_t res, const fmpq_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_fmpz_sub(ca_t res, const fmpz_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_ui_sub(ca_t res, ulong x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_si_sub(ca_t res, slong x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_sub(ca_t res, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_mul_fmpq(ca_t res, const ca_t x, const fmpq_t y, ca_ctx_t ctx) noexcept
    void ca_mul_fmpz(ca_t res, const ca_t x, const fmpz_t y, ca_ctx_t ctx) noexcept
    void ca_mul_ui(ca_t res, const ca_t x, ulong y, ca_ctx_t ctx) noexcept
    void ca_mul_si(ca_t res, const ca_t x, slong y, ca_ctx_t ctx) noexcept
    void ca_mul(ca_t res, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_inv(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_fmpq_div(ca_t res, const fmpq_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_fmpz_div(ca_t res, const fmpz_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_ui_div(ca_t res, ulong x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_si_div(ca_t res, slong x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_div_fmpq(ca_t res, const ca_t x, const fmpq_t y, ca_ctx_t ctx) noexcept
    void ca_div_fmpz(ca_t res, const ca_t x, const fmpz_t y, ca_ctx_t ctx) noexcept
    void ca_div_ui(ca_t res, const ca_t x, ulong y, ca_ctx_t ctx) noexcept
    void ca_div_si(ca_t res, const ca_t x, slong y, ca_ctx_t ctx) noexcept
    void ca_div(ca_t res, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_dot(ca_t res, const ca_t initial, int subtract, ca_srcptr x, slong xstep, ca_srcptr y, slong ystep, slong len, ca_ctx_t ctx) noexcept
    void ca_fmpz_poly_evaluate(ca_t res, const fmpz_poly_t poly, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_fmpq_poly_evaluate(ca_t res, const fmpq_poly_t poly, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_fmpz_mpoly_evaluate_horner(ca_t res, const fmpz_mpoly_t f, ca_srcptr x, const fmpz_mpoly_ctx_t mctx, ca_ctx_t ctx) noexcept
    void ca_fmpz_mpoly_evaluate_iter(ca_t res, const fmpz_mpoly_t f, ca_srcptr x, const fmpz_mpoly_ctx_t mctx, ca_ctx_t ctx) noexcept
    void ca_fmpz_mpoly_evaluate(ca_t res, const fmpz_mpoly_t f, ca_srcptr x, const fmpz_mpoly_ctx_t mctx, ca_ctx_t ctx) noexcept
    void ca_fmpz_mpoly_q_evaluate(ca_t res, const fmpz_mpoly_q_t f, ca_srcptr x, const fmpz_mpoly_ctx_t mctx, ca_ctx_t ctx) noexcept
    void ca_fmpz_mpoly_q_evaluate_no_division_by_zero(ca_t res, const fmpz_mpoly_q_t f, ca_srcptr x, const fmpz_mpoly_ctx_t mctx, ca_ctx_t ctx) noexcept
    void ca_inv_no_division_by_zero(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sqr(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_pow_fmpq(ca_t res, const ca_t x, const fmpq_t y, ca_ctx_t ctx) noexcept
    void ca_pow_fmpz(ca_t res, const ca_t x, const fmpz_t y, ca_ctx_t ctx) noexcept
    void ca_pow_ui(ca_t res, const ca_t x, ulong y, ca_ctx_t ctx) noexcept
    void ca_pow_si(ca_t res, const ca_t x, slong y, ca_ctx_t ctx) noexcept
    void ca_pow(ca_t res, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_pow_si_arithmetic(ca_t res, const ca_t x, slong n, ca_ctx_t ctx) noexcept
    void ca_sqrt_inert(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sqrt_nofactor(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sqrt_factor(ca_t res, const ca_t x, ulong flags, ca_ctx_t ctx) noexcept
    void ca_sqrt(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sqrt_ui(ca_t res, ulong n, ca_ctx_t ctx) noexcept
    void ca_abs(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sgn(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_csgn(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_arg(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_re(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_im(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_conj_deep(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_conj_shallow(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_conj(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_floor(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_ceil(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_exp(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_log(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sin_cos_exponential(ca_t res1, ca_t res2, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sin_cos_direct(ca_t res1, ca_t res2, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sin_cos_tangent(ca_t res1, ca_t res2, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sin_cos(ca_t res1, ca_t res2, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_sin(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_cos(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_tan_sine_cosine(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_tan_exponential(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_tan_direct(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_tan(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_cot(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_atan_logarithm(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_atan_direct(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_atan(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_asin_logarithm(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_acos_logarithm(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_asin_direct(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_acos_direct(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_asin(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_acos(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_gamma(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_erf(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_erfc(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_erfi(ca_t res, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_get_acb_raw(acb_t res, const ca_t x, slong prec, ca_ctx_t ctx) noexcept
    void ca_get_acb(acb_t res, const ca_t x, slong prec, ca_ctx_t ctx) noexcept
    void ca_get_acb_accurate_parts(acb_t res, const ca_t x, slong prec, ca_ctx_t ctx) noexcept
    char * ca_get_decimal_str(const ca_t x, slong digits, ulong flags, ca_ctx_t ctx) noexcept
    void ca_rewrite_complex_normal_form(ca_t res, const ca_t x, int deep, ca_ctx_t ctx) noexcept
    void ca_factor_init(ca_factor_t fac, ca_ctx_t ctx) noexcept
    void ca_factor_clear(ca_factor_t fac, ca_ctx_t ctx) noexcept
    void ca_factor_one(ca_factor_t fac, ca_ctx_t ctx) noexcept
    void ca_factor_print(const ca_factor_t fac, ca_ctx_t ctx) noexcept
    void ca_factor_insert(ca_factor_t fac, const ca_t base, const ca_t exp, ca_ctx_t ctx) noexcept
    void ca_factor_get_ca(ca_t res, const ca_factor_t fac, ca_ctx_t ctx) noexcept
    void ca_factor(ca_factor_t res, const ca_t x, ulong flags, ca_ctx_t ctx) noexcept
    void _ca_make_field_element(ca_t x, ca_field_srcptr new_index, ca_ctx_t ctx) noexcept
    void _ca_make_fmpq(ca_t x, ca_ctx_t ctx) noexcept



r"""
Deprecated module.

Functions were moved in arith_sage.pyx

TESTS::

    sage: from sage.libs.flint.arith import bell_number, bernoulli_number, euler_number, stirling_number_1, stirling_number_2, number_of_partitions, dedekind_sum, harmonic_number
    sage: bell_number(4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing bell_number from here is deprecated; please use "from sage.libs.flint.arith_sage import bell_number" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    15
    sage: bernoulli_number(4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing bernoulli_number from here is deprecated; please use "from sage.libs.flint.arith_sage import bernoulli_number" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    -1/30
    sage: euler_number(4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing euler_number from here is deprecated; please use "from sage.libs.flint.arith_sage import euler_number" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    5
    sage: stirling_number_1(2, 4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing stirling_number_1 from here is deprecated; please use "from sage.libs.flint.arith_sage import stirling_number_1" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    0
    sage: stirling_number_2(2, 4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing stirling_number_2 from here is deprecated; please use "from sage.libs.flint.arith_sage import stirling_number_2" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    0
    sage: number_of_partitions(4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing number_of_partitions from here is deprecated; please use "from sage.libs.flint.arith_sage import number_of_partitions" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    5
    sage: dedekind_sum(4, 5)
    doctest:warning
    ...
    DeprecationWarning:
    Importing dedekind_sum from here is deprecated; please use "from sage.libs.flint.arith_sage import dedekind_sum" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    -1/5
    sage: harmonic_number(4)
    doctest:warning
    ...
    DeprecationWarning:
    Importing harmonic_number from here is deprecated; please use "from sage.libs.flint.arith_sage import harmonic_number" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    25/12
"""

from sage.misc.lazy_import import LazyImport

bell_number = LazyImport('sage.libs.flint.arith_sage', 'bell_number', deprecation=36449)
bernoulli_number = LazyImport('sage.libs.flint.arith_sage', 'bernoulli_number', deprecation=36449)
euler_number = LazyImport('sage.libs.flint.arith_sage', 'euler_number', deprecation=36449)
stirling_number_1 = LazyImport('sage.libs.flint.arith_sage', 'stirling_number_1', deprecation=36449)
stirling_number_2 = LazyImport('sage.libs.flint.arith_sage', 'stirling_number_2', deprecation=36449)
number_of_partitions = LazyImport('sage.libs.flint.arith_sage', 'number_of_partitions', deprecation=36449)
dedekind_sum = LazyImport('sage.libs.flint.arith_sage', 'dedekind_sum', deprecation=36449)
harmonic_number = LazyImport('sage.libs.flint.arith_sage', 'harmonic_number', deprecation=36449)



# sage.doctest: needs sage.libs.gap
r"""
Hecke triangle groups

AUTHORS:

- Jonas Jermann (2013): initial version
"""

# ****************************************************************************
#       Copyright (C) 2013-2014 Jonas Jermann <jjermann2@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import divisors
from sage.groups.matrix_gps.finitely_generated import FinitelyGeneratedMatrixGroup_generic
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.misc.latex import latex
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.rings.infinity import infinity
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.structure.unique_representation import UniqueRepresentation

lazy_import("sage.functions.log", "exp")
lazy_import("sage.functions.gamma", "psi1")
lazy_import("sage.functions.trig", "sec")
lazy_import("sage.rings.number_field.number_field", "NumberField")
lazy_import("sage.rings.qqbar", ["AA", "AlgebraicField"])
lazy_import("sage.rings.universal_cyclotomic_field", "E")
lazy_import("sage.symbolic.constants", "pi")

from .hecke_triangle_group_element import HeckeTriangleGroupElement, cyclic_representative, coerce_AA


class HeckeTriangleGroup(FinitelyGeneratedMatrixGroup_generic,
                         UniqueRepresentation):
    r"""
    Hecke triangle group `(2, n, \infty)`.
    """

    Element = HeckeTriangleGroupElement

    @staticmethod
    def __classcall__(cls, n=3):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(QQ(3)) == HeckeTriangleGroup(int(3))
            True
        """
        if n == infinity:
            n = infinity
        else:
            n = ZZ(n)
            if n < 3:
                raise AttributeError("n has to be infinity or an Integer >= 3.")

        return super().__classcall__(cls, n)

    def __init__(self, n):
        r"""
        Hecke triangle group (2, n, infinity).

        Namely the von Dyck group corresponding to the triangle group
        with angles (pi/2, pi/n, 0).

        INPUT:

        - ``n`` -- ``infinity`` or an integer greater or equal to `3`

        OUTPUT: the Hecke triangle group for the given parameter `n`

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(12)
            sage: G
            Hecke triangle group for n = 12
            sage: G.category()
            Category of groups
        """

        self._n = n
        self.element_repr_method("default")

        if n in [3, infinity]:
            self._base_ring = ZZ
            self._lam = ZZ.one() if n == 3 else ZZ(2)
        else:
            lam_symbolic = coerce_AA(E(2 * n) + ~E(2 * n))
            K = NumberField(self.lam_minpoly(), 'lam',
                            embedding=lam_symbolic)
            # self._base_ring = K.order(K.gens())
            self._base_ring = K.maximal_order()
            self._lam = self._base_ring.gen(1)

        T = matrix(self._base_ring, [[1, self._lam], [0, 1]])
        S = matrix(self._base_ring, [[0, -1], [1, 0]])

        FinitelyGeneratedMatrixGroup_generic.__init__(self, ZZ(2), self._base_ring, [S, T])

    def _repr_(self):
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(10)
            Hecke triangle group for n = 10
        """
        return f"Hecke triangle group for n = {self._n}"

    def _latex_(self):
        r"""
        Return the LaTeX representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: a = HeckeTriangleGroup(5)
            sage: latex(a)
            \Gamma^{(5)}
        """
        return '\\Gamma^{(%s)}' % latex(self._n)

    def element_repr_method(self, method=None):
        r"""
        Either return or set the representation method for elements of ``self``.

        INPUT:

        - ``method`` -- if ``method=None`` (default) the current default
          representation method is returned. Otherwise the default method is
          set to ``method``. If ``method`` is not available a
          :exc:`ValueError` is raised. Possible methods are:

          - ``default``: use the usual representation method for matrix group
            elements

          - ``basic``: the representation is given as a word in ``S`` and
            powers of ``T``

          - ``conj``: the conjugacy representative of the element is
            represented as a word in powers of the basic blocks, together with
            an unspecified conjugation matrix

          - ``block``: same as ``conj`` but the conjugation matrix is
            specified as well

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(5)
            sage: G.element_repr_method()
            'default'
            sage: G.element_repr_method("basic")
            sage: G.element_repr_method()
            'basic'
        """
        if method is None:
            return self._element_repr_method
        if method in ["default", "basic", "block", "conj"]:
            self._element_repr_method = method
        else:
            raise ValueError(f"the specified method {method} is not supported")

    def one(self):
        r"""
        Return the identity element/matrix for ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(10)
            sage: G(1) == G.one()
            True
            sage: G(1)
            [1 0]
            [0 1]

            sage: G(1).parent()
            Hecke triangle group for n = 10
        """
        return self.I()

    def lam_minpoly(self):
        r"""
        Return the minimal polynomial of the corresponding lambda parameter of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(10).lam_minpoly()
            x^4 - 5*x^2 + 5
            sage: HeckeTriangleGroup(17).lam_minpoly()
            x^8 - x^7 - 7*x^6 + 6*x^5 + 15*x^4 - 10*x^3 - 10*x^2 + 4*x + 1
            sage: HeckeTriangleGroup(infinity).lam_minpoly()
            x - 2
        """
        # TODO: Write an explicit (faster) implementation
        n = self._n
        if n == infinity:
            lam_symbolic = QQ(2)
        else:
            lam_symbolic = E(2 * n) + ~E(2 * n)
        return lam_symbolic.minpoly()

    def base_ring(self):
        r"""
        Return the base ring of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(n=infinity).base_ring()
            Integer Ring
            sage: HeckeTriangleGroup(n=7).base_ring()
            Maximal Order generated by lam in Number Field in lam with defining polynomial x^3 - x^2 - 2*x + 1 with lam = 1.801937735804839?
        """
        return self._base_ring

    def base_field(self):
        r"""
        Return the base field of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(n=infinity).base_field()
            Rational Field
            sage: HeckeTriangleGroup(n=7).base_field()
            Number Field in lam with defining polynomial x^3 - x^2 - 2*x + 1 with lam = 1.801937735804839?
        """
        if self._n in [3, infinity]:
            return QQ
        return self._base_ring.number_field()

    def n(self):
        r"""
        Return the parameter ``n`` of ``self``, where
        ``pi/n`` is the angle at ``rho`` of the corresponding
        basic hyperbolic triangle with vertices ``i``, ``rho``
        and ``infinity``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(10).n()
            10
            sage: HeckeTriangleGroup(infinity).n()
            +Infinity
        """
        return self._n

    def lam(self):
        r"""
        Return the parameter ``lambda`` of ``self``,
        where ``lambda`` is twice the real part of ``rho``,
        lying between ``1`` (when ``n=3``) and ``2`` (when ``n=infinity``).

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).lam()
            1
            sage: HeckeTriangleGroup(4).lam()
            lam
            sage: HeckeTriangleGroup(4).lam()^2
            2
            sage: HeckeTriangleGroup(6).lam()^2
            3
            sage: AA(HeckeTriangleGroup(10).lam())
            1.9021130325903...?
            sage: HeckeTriangleGroup(infinity).lam()
            2
        """
        return self._lam

    def rho(self):
        r"""
        Return the vertex ``rho`` of the basic hyperbolic
        triangle which describes ``self``. ``rho`` has
        absolute value 1 and angle ``pi/n``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).rho() == QQbar(1/2 + sqrt(3)/2*i)
            True
            sage: HeckeTriangleGroup(4).rho() == QQbar(sqrt(2)/2*(1 + i))
            True
            sage: HeckeTriangleGroup(6).rho() == QQbar(sqrt(3)/2 + 1/2*i)
            True
            sage: HeckeTriangleGroup(10).rho()
            0.95105651629515...? + 0.30901699437494...?*I
            sage: HeckeTriangleGroup(infinity).rho()
            1
        """
        # TODO: maybe rho should be replaced by -rhobar
        # Also we could use NumberFields...
        if self._n == infinity:
            return AA.one()
        rho = AlgebraicField()(E(2 * self._n))
        rho.simplify()
        return rho

    def alpha(self):
        r"""
        Return the parameter ``alpha`` of ``self``.

        This is the first parameter of the hypergeometric series used
        in the calculation of the Hauptmodul of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).alpha()
            1/12
            sage: HeckeTriangleGroup(4).alpha()
            1/8
            sage: HeckeTriangleGroup(5).alpha()
            3/20
            sage: HeckeTriangleGroup(6).alpha()
            1/6
            sage: HeckeTriangleGroup(10).alpha()
            1/5
            sage: HeckeTriangleGroup(infinity).alpha()
            1/4
        """
        return QQ((1, 2)) * (QQ((1, 2)) - ZZ.one() / self._n)

    def beta(self):
        r"""
        Return the parameter ``beta`` of ``self``.

        This is the second parameter of the hypergeometric series used
        in the calculation of the Hauptmodul of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).beta()
            5/12
            sage: HeckeTriangleGroup(4).beta()
            3/8
            sage: HeckeTriangleGroup(5).beta()
            7/20
            sage: HeckeTriangleGroup(6).beta()
            1/3
            sage: HeckeTriangleGroup(10).beta()
            3/10
            sage: HeckeTriangleGroup(infinity).beta()
            1/4
        """
        return QQ((1, 2)) * (QQ((1, 2)) + ZZ.one() / self._n)

    # We use cached method here to create unique instances of basic matrices
    # (major performance gain)
    @cached_method
    def I(self):
        r"""
        Return the identity element/matrix for ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(10).I()
            [1 0]
            [0 1]

            sage: HeckeTriangleGroup(10).I().parent()
            Hecke triangle group for n = 10
        """
        return self(matrix(self._base_ring, [[1, 0], [0, 1]]), check=False)

    # We use cached method here to create unique instances of basic matrices
    # (major performance gain)
    @cached_method
    def T(self, m=1):
        r"""
        Return the element in ``self`` corresponding to the translation by
        ``m*self.lam()``.

        INPUT:

        - ``m`` -- integer (default: 1); namely the second generator of
          ``self``

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).T()
            [1 1]
            [0 1]
            sage: HeckeTriangleGroup(10).T(-4)
            [     1 -4*lam]
            [     0      1]
            sage: HeckeTriangleGroup(10).T().parent()
            Hecke triangle group for n = 10
        """
        return self(matrix(self._base_ring, [[1, self._lam * m], [0, 1]]),
                    check=False)

    # We use cached method here to create unique instances of basic matrices
    # (major performance gain)
    @cached_method
    def S(self):
        r"""
        Return the generator of ``self`` corresponding to the
        conformal circle inversion.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).S()
            [ 0 -1]
            [ 1  0]
            sage: HeckeTriangleGroup(10).S()
            [ 0 -1]
            [ 1  0]
            sage: HeckeTriangleGroup(10).S()^2 == -HeckeTriangleGroup(10).I()
            True
            sage: HeckeTriangleGroup(10).S()^4 == HeckeTriangleGroup(10).I()
            True

            sage: HeckeTriangleGroup(10).S().parent()
            Hecke triangle group for n = 10
        """
        return self.gen(0)

    # We use cached method here to create unique instances of basic matrices
    # (major performance gain)
    @cached_method
    def U(self):
        r"""
        Return an alternative generator of ``self`` instead of ``T``.
        ``U`` stabilizes ``rho`` and has order ``2*self.n()``.

        If ``n=infinity`` then ``U`` is parabolic and has infinite order,
        it then fixes the cusp ``[-1]``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).U()
            [ 1 -1]
            [ 1  0]
            sage: HeckeTriangleGroup(3).U()^3 == -HeckeTriangleGroup(3).I()
            True
            sage: HeckeTriangleGroup(3).U()^6 == HeckeTriangleGroup(3).I()
            True
            sage: HeckeTriangleGroup(10).U()
            [lam  -1]
            [  1   0]
            sage: HeckeTriangleGroup(10).U()^10 == -HeckeTriangleGroup(10).I()
            True
            sage: HeckeTriangleGroup(10).U()^20 == HeckeTriangleGroup(10).I()
            True

            sage: HeckeTriangleGroup(10).U().parent()
            Hecke triangle group for n = 10
        """
        return self.T() * self.S()

    def V(self, j):
        r"""
        Return the j'th generator for the usual representatives of
        conjugacy classes of ``self``. It is given by ``V=U^(j-1)*T``.

        INPUT:

        - ``j`` --integer; to get the usual representatives
          ``j`` should range from ``1`` to ``self.n()-1``

        OUTPUT: the corresponding matrix/element

        The matrix is parabolic if ``j`` is congruent to `\pm 1` modulo ``self.n()``.
        It is elliptic if ``j`` is congruent to 0 modulo ``self.n()``.
        It is hyperbolic otherwise.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(3)
            sage: G.V(0) == -G.S()
            True
            sage: G.V(1) == G.T()
            True
            sage: G.V(2)
            [1 0]
            [1 1]
            sage: G.V(3) == G.S()
            True

            sage: G = HeckeTriangleGroup(5)
            sage: G.element_repr_method("default")
            sage: G.V(1)
            [  1 lam]
            [  0   1]
            sage: G.V(2)
            [lam lam]
            [  1 lam]
            sage: G.V(3)
            [lam   1]
            [lam lam]
            sage: G.V(4)
            [  1   0]
            [lam   1]
            sage: G.V(5) == G.S()
            True
        """
        return self.U()**(j-1) * self.T()

    def dvalue(self):
        r"""
        Return a symbolic expression (or an exact value in case n=3, 4, 6)
        for the transfinite diameter (or capacity) of ``self``.

        This is the first nontrivial Fourier coefficient of the Hauptmodul
        for the Hecke triangle group in case it is normalized to ``J_inv(i)=1``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).dvalue()
            1/1728
            sage: HeckeTriangleGroup(4).dvalue()
            1/256
            sage: HeckeTriangleGroup(5).dvalue()
            e^(2*euler_gamma - 4*pi/(sqrt(5) + 1) + psi(17/20) + psi(13/20))
            sage: HeckeTriangleGroup(6).dvalue()
            1/108
            sage: HeckeTriangleGroup(10).dvalue()
            e^(2*euler_gamma - 4*pi/sqrt(2*sqrt(5) + 10) + psi(4/5) + psi(7/10))
            sage: HeckeTriangleGroup(infinity).dvalue()
            1/64
        """
        n = self._n
        if n == 3:
            return 1 / ZZ(2**6*3**3)
        if n == 4:
            return 1 / ZZ(2**8)
        if n == 6:
            return 1 / ZZ(2**2*3**3)
        if n == infinity:
            return 1 / ZZ(2**6)
        # reference for this formula ?
        return exp(-ZZ(2)*psi1(ZZ.one()) + psi1(ZZ.one()-self.alpha())
                   + psi1(ZZ.one()-self.beta()) - pi*sec(pi/self._n))

    def is_arithmetic(self) -> bool:
        r"""
        Return ``True`` if ``self`` is an arithmetic subgroup.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(3).is_arithmetic()
            True
            sage: HeckeTriangleGroup(4).is_arithmetic()
            True
            sage: HeckeTriangleGroup(5).is_arithmetic()
            False
            sage: HeckeTriangleGroup(6).is_arithmetic()
            True
            sage: HeckeTriangleGroup(10).is_arithmetic()
            False
            sage: HeckeTriangleGroup(infinity).is_arithmetic()
            True
        """
        return self._n in [ZZ(3), ZZ(4), ZZ(6), infinity]

    def get_FD(self, z):
        r"""
        Return a tuple (A,w) which determines how to map ``z``
        to the usual (strict) fundamental domain of ``self``.

        INPUT:

        - ``z`` -- a complex number or an element of AlgebraicField()

        OUTPUT:

        A tuple ``(A, w)``.

        - ``A`` -- a matrix in ``self`` such that ``A.acton(w)==z``
          (if ``z`` is exact at least)

        - ``w`` -- a complex number or an element of AlgebraicField()
          (depending on the type ``z``) which lies inside the (strict)
          fundamental domain of ``self`` (``self.in_FD(w)==True``) and
          which is equivalent to ``z`` (by the above property)

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(8)
            sage: z = AlgebraicField()(1+i/2)
            sage: A, w = G.get_FD(z)
            sage: A
            [-lam    1]
            [  -1    0]
            sage: A.acton(w) == z
            True

            sage: from sage.modular.modform_hecketriangle.space import ModularForms
            sage: z = (134.12 + 0.22*i).n()
            sage: A, w = G.get_FD(z)
            sage: A
            [-73*lam^3 + 74*lam       73*lam^2 - 1]
            [        -lam^2 + 1                lam]
            sage: w
            0.769070776942... + 0.779859114103...*I
            sage: z
            134.120000000... + 0.220000000000...*I
            sage: A.acton(w)
            134.1200000... + 0.2200000000...*I
        """
        ID = self.I()
        T = self.T()
        S = self.S()
        TI = self.T(-1)

        A = ID
        w = z
        while abs(w) < ZZ.one() or abs(w.real()) > self.lam() / ZZ(2):
            if abs(w) < ZZ.one():
                w = self.S().acton(w)
                A = S*A
            while w.real() >= self.lam() / ZZ(2):
                w = TI.acton(w)
                A = TI*A
            while w.real() < -self.lam() / ZZ(2):
                w = T.acton(w)
                A = T*A
        if w.real() == self.lam() / ZZ(2):
            w = TI.acton(w)
            A = TI*A
        if abs(w) == ZZ.one() and w.real() > ZZ.zero():
            w = S.acton(w)
            A = S*A

        AI = A.inverse()

        return (AI, A.acton(z))

    def in_FD(self, z) -> bool:
        r"""
        Return ``True`` if ``z`` lies in the (strict) fundamental
        domain of ``self``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: HeckeTriangleGroup(5).in_FD(CC(1.5/2 + 0.9*i))
            True
            sage: HeckeTriangleGroup(4).in_FD(CC(1.5/2 + 0.9*i))
            False
        """
        return self.get_FD(z)[0] == self.I()

    def root_extension_field(self, D):
        r"""
        Return the quadratic extension field of the base field by
        the square root of the given discriminant ``D``.

        INPUT:

        - ``D`` -- an element of the base ring of ``self``
          corresponding to a discriminant

        OUTPUT:

        A relative (at most quadratic) extension to the base field
        of ``self`` in the variable ``e`` which corresponds to ``sqrt(D)``.
        If the extension degree is ``1`` then the base field is returned.

        The correct embedding is the positive resp. positive imaginary one.
        Unfortunately no default embedding can be specified for relative
        number fields yet.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=infinity)
            sage: G.root_extension_field(32)
            Number Field in e with defining polynomial x^2 - 32
            sage: G.root_extension_field(-4)
            Number Field in e with defining polynomial x^2 + 4
            sage: G.root_extension_field(4) == G.base_field()
            True

            sage: G = HeckeTriangleGroup(n=7)
            sage: lam = G.lam()
            sage: D = 4*lam^2 + 4*lam - 4
            sage: G.root_extension_field(D)
            Number Field in e with defining polynomial x^2 - 4*lam^2 - 4*lam + 4 over its base field
            sage: G.root_extension_field(4) == G.base_field()
            True
            sage: D = lam^2 - 4
            sage: G.root_extension_field(D)
            Number Field in e with defining polynomial x^2 - lam^2 + 4 over its base field
        """
        K = self.base_field()
        x = PolynomialRing(K, 'x').gen()
        D = self.base_ring()(D)

        if D.is_square():
            return K
        # unfortunately we can't set embeddings for relative extensions :-(
        # return K.extension(x**2 - D, 'e', embedding=AA(D).sqrt())

        L = K.extension(x**2 - D, 'e')

        # e = AA(D).sqrt()
        # emb = L.hom([e])
        # L._unset_embedding()
        # L.register_embedding(emb)

        # return NumberField(L.absolute_polynomial(), 'e', structure=AbsoluteFromRelative(L), embedding=(???))
        return L

    # We cache this method for performance reasons (it is repeatedly reused)
    @cached_method
    def root_extension_embedding(self, D, K=None):
        r"""
        Return the correct embedding from the root extension field
        of the given discriminant ``D``  to the field ``K``.

        Also see the method ``root_extension_embedding(K)`` of
        ``HeckeTriangleGroupElement`` for more examples.

        INPUT:

        - ``D`` -- an element of the base ring of ``self``
          corresponding to a discriminant

        - ``K`` -- a field to which we want the (correct) embedding;
          if ``K=None`` (default) then ``AlgebraicField()`` is
          used for positive ``D`` and ``AlgebraicRealField()``
          otherwise

        OUTPUT:

        The corresponding embedding if it was found.
        Otherwise a :exc:`ValueError` is raised.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=infinity)
            sage: G.root_extension_embedding(32)
            Ring morphism:
              From: Number Field in e with defining polynomial x^2 - 32
              To:   Algebraic Real Field
              Defn: e |--> 5.656854249492...?
            sage: G.root_extension_embedding(-4)
            Ring morphism:
              From: Number Field in e with defining polynomial x^2 + 4
              To:   Algebraic Field
              Defn: e |--> 2*I
            sage: G.root_extension_embedding(4)
            Coercion map:
              From: Rational Field
              To:   Algebraic Real Field

            sage: G = HeckeTriangleGroup(n=7)
            sage: lam = G.lam()
            sage: D = 4*lam^2 + 4*lam - 4
            sage: G.root_extension_embedding(D, CC)
            Relative number field morphism:
              From: Number Field in e with defining polynomial x^2 - 4*lam^2 - 4*lam + 4 over its base field
              To:   Complex Field with 53 bits of precision
              Defn: e |--> 4.02438434522...
                    lam |--> 1.80193773580...
            sage: D = lam^2 - 4
            sage: G.root_extension_embedding(D)
            Relative number field morphism:
              From: Number Field in e with defining polynomial x^2 - lam^2 + 4 over its base field
              To:   Algebraic Field
              Defn: e |--> 0.?... + 0.867767478235...?*I
                    lam |--> 1.801937735804...?
        """
        D = self.base_ring()(D)
        F = self.root_extension_field(D)
        if K is None:
            if coerce_AA(D) > 0:
                K = AA
            else:
                K = AlgebraicField()

        L = list(F.embeddings(K))

        # Three possibilities up to numerical artefacts:
        # (1) emb = e, purely imaginary
        # (2) emb = e or lam (can't distinguish), purely real
        # (3) emb = (e,lam), e purely imaginary, lam purely real
        # (4) emb = (e,lam), e purely real, lam purely real
        # There always exists one emb with "e" positive resp. positive imaginary
        # and if there is a lam there exists a positive one...
        #
        # Criteria to pick the correct "maximum":
        # 1. First figure out if e resp. lam is purely real or imaginary
        #    (using "abs(e.imag()) > abs(e.real())")
        # 2. In the purely imaginary case we don't want anything negative imaginary
        #    and we know the positive case is unique after sorting lam
        # 3. For the remaining cases we want the biggest real part
        #    (and lam should get comparison priority)
        def emb_key(emb):
            L = []
            gens_len = len(emb.im_gens())
            for k in range(gens_len):
                a = emb.im_gens()[k]
                try:
                    a.simplify()
                    a.exactify()
                except AttributeError:
                    pass
                # If a is purely imaginary:
                if abs(a.imag()) > abs(a.real()):
                    if a.imag() < 0:
                        a = -infinity
                    else:
                        a = ZZ(0)
                else:
                    a = a.real()

                L.append(a)

            L.reverse()
            return L

        if len(L) > 1:
            L.sort(key=emb_key)
        return L[-1]

    def _elliptic_conj_reps(self):
        r"""
        Store all elliptic conjugacy representatives in the
        internal dictionary. This is a helper function
        used by :meth:`_conjugacy_representatives`.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=5)
            sage: G.element_repr_method("conj")
            sage: G._elliptic_conj_reps()
            sage: sorted(G._conj_prim.items())
            [(-4, [[S], [S]]), (lam - 3, [[U], [U]]), (0, [[V(4)]])]
            sage: sorted(G._conj_nonprim.items())
            [(-lam - 2, [[U^(-2)], [U^2], [U^(-2)], [U^2]]), (lam - 3, [[U^(-1)], [U^(-1)]])]
        """
        if not hasattr(self, "_max_block_length"):
            self._conjugacy_representatives()
        elif ZZ(-4) in self._conj_prim:
            return

        D = self.U().discriminant()
        if D not in self._conj_prim:
            self._conj_prim[D] = []
        self._conj_prim[D].append(self.U())

        D = self.S().discriminant()
        if D not in self._conj_prim:
            self._conj_prim[D] = []
        self._conj_prim[D].append(self.S())

        other_reps = [self.U()**k
                      for k in range(-((self.n() - 1) / 2).floor(),
                                     self.n() // 2 + 1)
                      if k not in [0, 1]]

        for v in other_reps:
            D = v.discriminant()
            if D not in self._conj_nonprim:
                self._conj_nonprim[D] = []
            self._conj_nonprim[D].append(v)

    def _conjugacy_representatives(self, max_block_length=0, D=None):
        r"""
        Store conjugacy representatives up to block length
        ``max_block_length`` (a nonnegative integer, default: 0)
        in the internal dictionary. Previously calculated data is reused.
        This is a helper function for e.g. :meth:`class_number`.

        The set of all (hyperbolic) conjugacy types of block length
        ``t`` is stored in ``self._conj_block[t]``.
        The set of all primitive representatives (so far) with
        discriminant ``D`` is stored in ``self._conj_prim[D]``.
        The set of all non-primitive representatives (so far) with
        discriminant ``D`` is stored in ``self._conj_nonprim[D]``.

        The case of nonpositive discriminants is done manually.

        INPUT:

        - ``max_block_length`` -- nonnegative integer (default: `0`); the
          maximal block length

        - ``D`` -- an element/discriminant of the base ring or more generally
          an upper bound for the involved discriminants. If ``D != None`` then
          an upper bound for ``max_block_length`` is deduced from ``D``
          (default: ``None``).

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=5)
            sage: G.element_repr_method("conj")
            sage: G._conjugacy_representatives(2)

            sage: sorted(G._conj_block[2])
            [((2, 1), (1, 1)),
             ((2, 2),),
             ((3, 1), (1, 1)),
             ((3, 1), (2, 1)),
             ((3, 2),),
             ((4, 1), (1, 1)),
             ((4, 1), (2, 1)),
             ((4, 1), (3, 1))]

            sage: [key for key in sorted(G._conj_prim)]
            [-4, lam - 3, 0, 4*lam, 7*lam + 6, 9*lam + 5, 15*lam + 6, 33*lam + 21]
            sage: for key in sorted(G._conj_prim):
            ....:     print(sorted(G._conj_prim[key]))
            [[S], [S]]
            [[U], [U]]
            [[V(4)]]
            [[V(3)], [V(2)]]
            [[V(1)*V(4)]]
            [[V(3)*V(4)], [V(1)*V(2)]]
            [[V(2)*V(4)], [V(1)*V(3)]]
            [[V(2)*V(3)]]
            sage: [key for key in sorted(G._conj_nonprim)]
            [-lam - 2, lam - 3, 32*lam + 16]

            sage: for key in sorted(G._conj_nonprim):
            ....:     print(sorted(G._conj_nonprim[key]))
            [[U^(-2)], [U^(-2)], [U^2], [U^2]]
            [[U^(-1)], [U^(-1)]]
            [[V(3)^2], [V(2)^2]]

            sage: G.element_repr_method("default")
        """
        from sage.combinat.partition import OrderedPartitions
        from sage.combinat.combinat import tuples

        if D is not None:
            max_block_length = max(AA.zero(),
                                   coerce_AA((D + 4)/(self.lam()**2))).sqrt().floor()
        else:
            try:
                max_block_length = ZZ(max_block_length)
                if max_block_length < 0:
                    raise TypeError
            except TypeError:
                raise ValueError("max_block_length must be a nonnegative integer!")

        if not hasattr(self, "_max_block_length"):
            self._max_block_length = ZZ.zero()
            self._conj_block = {}
            self._conj_nonprim = {}
            self._conj_prim = {}

            # It is not clear how to define the class number for D=0:
            # Conjugacy classes are V(n-1)^(+-k) for arbitrary k
            # and the trivial class (what about self_conj_block[0]?).
            #
            # One way is to define it using the fixed points and in
            # that case V(n-1) would be a good representative.
            # The non-primitive case is unclear however...
            #
            # We set it here to ensure that 0 is enlisted as a discriminant...
            #
            self._conj_prim[ZZ.zero()] = []
            self._conj_prim[ZZ.zero()].append(self.V(self.n()-1))

            self._elliptic_conj_reps()

        if max_block_length <= self._max_block_length:
            return

        def is_cycle(seq) -> bool:
            length = len(seq)
            for n in divisors(length):
                if n < length and is_cycle_of_length(seq, n):
                    return True
            return False

        def is_cycle_of_length(seq, n) -> bool:
            return all(seq[j] == seq[j % n] for j in range(n, len(seq)))

        j_list = range(1, self.n())

        for t in range(self._max_block_length + 1, max_block_length + 1):
            t_ZZ = ZZ(t)
            if t_ZZ not in self._conj_block:
                self._conj_block[t_ZZ] = set()

            partitions = OrderedPartitions(t).list()
            for par in partitions:
                len_par = len(par)
                exp_list = tuples(j_list, len_par)
                for ex in exp_list:
                    keep = True
                    if len_par > 1:
                        for k in range(-1, len_par - 1):
                            if ex[k] == ex[k + 1]:
                                keep = False
                                break
                    # We don't want powers of V(1)
                    elif ex[0] == 1:
                        keep = False
                    # But: Do we maybe want powers of V(n-1)??
                    # For now we exclude the parabolic cases...
                    elif ex[0] == self.n() - 1:
                        keep = False

                    if keep:
                        conj_type = cyclic_representative(tuple((ZZ(ex[k]), ZZ(par[k])) for k in range(len_par)))
                        self._conj_block[t_ZZ].add(conj_type)

            for el in self._conj_block[t_ZZ]:
                group_el = prod([self.V(el[k][0])**el[k][1] for k in range(len(el))])

                D = group_el.discriminant()
                assert coerce_AA(D) > 0

                # The primitive cases
                # if group_el.is_primitive():
                if not ((len(el) == 1 and el[0][1] > 1) or is_cycle(el)):
                    if D not in self._conj_prim:
                        self._conj_prim[D] = []
                    self._conj_prim[D].append(group_el)
                # The remaining cases
                else:
                    if D not in self._conj_nonprim:
                        self._conj_nonprim[D] = []
                    self._conj_nonprim[D].append(group_el)

        self._max_block_length = max_block_length

    def class_representatives(self, D, primitive=True):
        r"""
        Return a representative for each conjugacy class for the
        discriminant ``D`` (ignoring the sign).

        If ``primitive=True`` only one representative for each
        fixed point is returned (ignoring sign).

        INPUT:

        - ``D`` -- an element of the base ring corresponding
          to a valid discriminant

        - ``primitive`` -- if ``True`` (default) then only primitive
          representatives are considered

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)
            sage: G.element_repr_method("conj")

            sage: R = G.class_representatives(4)
            sage: R
            [[V(2)]]
            sage: [v.continued_fraction()[1] for v in R]
            [(2,)]

            sage: R = G.class_representatives(0)
            sage: R
            [[V(3)]]
            sage: [v.continued_fraction()[1] for v in R]
            [(1, 2)]

            sage: R = G.class_representatives(-4)
            sage: R
            [[S]]
            sage: R = G.class_representatives(-4, primitive=False)
            sage: R
            [[S], [U^2]]

            sage: R = G.class_representatives(G.lam()^2 - 4)
            sage: R
            [[U]]
            sage: R = G.class_representatives(G.lam()^2 - 4, primitive=False)
            sage: R
            [[U], [U^(-1)]]

            sage: R = G.class_representatives(14)
            sage: sorted(R)
            [[V(2)*V(3)], [V(1)*V(2)]]
            sage: sorted(v.continued_fraction()[1] for v in R)
            [(1, 2, 2), (3,)]

            sage: R = G.class_representatives(32)
            sage: sorted(R)
            [[V(3)^2*V(1)], [V(1)^2*V(3)]]
            sage: [v.continued_fraction()[1] for v in sorted(R)]
            [(1, 2, 1, 3), (1, 4)]

            sage: R = G.class_representatives(32, primitive=False)
            sage: sorted(R)
            [[V(3)^2*V(1)], [V(1)^2*V(3)], [V(2)^2]]

            sage: G.element_repr_method("default")
        """
        if coerce_AA(D) == 0 and not primitive:
            raise ValueError("There are infinitely many non-primitive conjugacy classes of discriminant 0.")

        self._conjugacy_representatives(D=D)

        L = []
        if D in self._conj_prim:
            L += self._conj_prim[D]
        if not primitive and D in self._conj_nonprim:
            L += self._conj_nonprim[D]

        if not L:
            raise ValueError("D = {} is not a{} discriminant for {}".format(D, " primitive" if primitive else "", self))
        else:
            return L

    def class_number(self, D, primitive=True):
        r"""
        Return the class number of ``self`` for the discriminant ``D``.

        This is the number of conjugacy classes of (primitive) elements
        of discriminant ``D``.

        Note: Due to the 1-1 correspondence with hyperbolic fixed
        points resp. hyperbolic binary quadratic forms
        this also gives the class number in those cases.

        INPUT:

        - ``D`` -- an element of the base ring corresponding
          to a valid discriminant

        - ``primitive`` -- if ``True`` (default) then only primitive
          elements are considered

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)

            sage: G.class_number(4)
            1
            sage: G.class_number(4, primitive=False)
            1
            sage: G.class_number(14)
            2
            sage: G.class_number(32)
            2
            sage: G.class_number(32, primitive=False)
            3
            sage: G.class_number(68)
            4
        """

        if coerce_AA(D) <= 0:
            raise NotImplementedError

        self._conjugacy_representatives(D=D)

        num = ZZ(0)
        if D in self._conj_prim:
            num = len(self._conj_prim[D])
        if not primitive and D in self._conj_nonprim:
            num += len(self._conj_nonprim[D])

        if num == 0:
            raise ValueError("D = {} is not a{} discriminant for {}".format(D, " primitive" if primitive else "", self))
        else:
            return num

    def is_discriminant(self, D, primitive=True) -> bool:
        r"""
        Return whether ``D`` is a discriminant of an element of ``self``.

        Note: Checking that something isn't a discriminant takes much
        longer than checking for valid discriminants.

        INPUT:

        - ``D`` -- an element of the base ring

        - ``primitive`` -- if ``True`` (default) then only primitive
          elements are considered

        OUTPUT:

        ``True`` if ``D`` is a primitive discriminant (a discriminant of
        a primitive element) and ``False`` otherwise.
        If ``primitive=False`` then also non-primitive elements are considered.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)

            sage: G.is_discriminant(68)
            True
            sage: G.is_discriminant(196, primitive=False)    # long time
            True
            sage: G.is_discriminant(2)
            False
        """
        self._conjugacy_representatives(0)
        t_bound = max(AA.zero(),
                      coerce_AA((D + 4) / (self.lam()**2))).sqrt().floor()
        for t in range(self._max_block_length + 1, t_bound + 1):
            self._conjugacy_representatives(t)

            if D in self._conj_prim:
                return True
            if not primitive and D in self._conj_nonprim:
                return True

        if D in self._conj_prim:
            return True
        return not primitive and D in self._conj_nonprim

    def list_discriminants(self, D, primitive=True, hyperbolic=True, incomplete=False):
        r"""
        Return a list of all discriminants up to some upper bound ``D``.

        INPUT:

        - ``D`` -- an element/discriminant of the base ring or
          more generally an upper bound for the discriminant

        - ``primitive`` -- if ``True`` (default) then only primitive
          discriminants are listed

        - ``hyperbolic`` -- if ``True`` (default) then only positive
          discriminants are listed

        - ``incomplete`` -- if ``True`` (default: ``False``) then all
          (also higher) discriminants which were gathered so far are listed
          (however there might be missing discriminants in between).

        OUTPUT: list of discriminants less than or equal to ``D``

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)
            sage: G.list_discriminants(D=68)
            [4, 12, 14, 28, 32, 46, 60, 68]
            sage: G.list_discriminants(D=0, hyperbolic=False, primitive=False)
            [-4, -2, 0]

            sage: G = HeckeTriangleGroup(n=5)
            sage: G.list_discriminants(D=20)
            [4*lam, 7*lam + 6, 9*lam + 5]
            sage: G.list_discriminants(D=0, hyperbolic=False, primitive=False)
            [-4, -lam - 2, lam - 3, 0]
        """

        self._conjugacy_representatives(D=D)

        if incomplete:
            max_D = infinity
        else:
            max_D = coerce_AA(D)

        if hyperbolic:
            L = [key for key in self._conj_prim if 0 < coerce_AA(key) <= max_D]
        else:
            L = [key for key in self._conj_prim if coerce_AA(key) <= max_D]

        if not primitive:
            if hyperbolic:
                L += [key for key in self._conj_nonprim
                      if 0 < coerce_AA(key) <= max_D and key not in L]
            else:
                L += [key for key in self._conj_nonprim
                      if coerce_AA(key) <= max_D and key not in L]

        return sorted(L, key=coerce_AA)

    # TODO: non-primitive ones?
    def reduced_elements(self, D):
        r"""
        Return all reduced (primitive) elements of discriminant ``D``.

        Also see the element method :meth:`is_reduced` for more information.

        - ``D`` -- an element of the base ring corresponding
          to a valid discriminant

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)
            sage: R = G.reduced_elements(D=12)
            sage: R
            [
            [    5  -lam]  [     5 -3*lam]
            [3*lam    -1], [   lam     -1]
            ]
            sage: [v.continued_fraction() for v in R]
            [((), (1, 3)), ((), (3, 1))]
            sage: R = G.reduced_elements(D=14)
            sage: sorted(R)
            [
            [3*lam    -1]  [4*lam    -3]  [ 5*lam     -7]  [ 5*lam     -3]
            [    1     0], [    3  -lam], [     3 -2*lam], [     7 -2*lam]
            ]
            sage: sorted(v.continued_fraction() for v in R)
            [((), (1, 2, 2)), ((), (2, 1, 2)), ((), (2, 2, 1)), ((), (3,))]
        """
        L = self.class_representatives(D=D, primitive=True)
        R = []
        for v in L:
            R += v.reduced_elements()

        return R

    def simple_elements(self, D):
        r"""
        Return all simple elements of discriminant ``D``.

        Also see the element method ``is_simple()`` for more information.

        - ``D`` -- an element of the base ring corresponding to a valid
          discriminant

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)
            sage: sorted(G.simple_elements(D=12))
            [
            [  1 lam]  [  3 lam]
            [lam   3], [lam   1]
            ]
            sage: sorted(G.simple_elements(D=14))
            [
            [  lam     1]  [  lam     3]  [2*lam     1]  [2*lam     3]
            [    3 2*lam], [    1 2*lam], [    3   lam], [    1   lam]
            ]
        """
        L = self.class_representatives(D=D, primitive=True)
        R = []
        for v in L:
            R += v.simple_elements()

        return R

    def rational_period_functions(self, k, D):
        r"""
        Return a list of basic rational period functions of weight ``k`` for
        discriminant ``D``.

        The list is expected to be a generating set for all rational
        period functions of the given weight and discriminant (unknown).

        The method assumes that ``D > 0``.

        Also see the element method ``rational_period_function`` for more information.

        - ``k`` -- even integer, the desired weight of the rational period
          functions

        - ``D`` -- an element of the base ring corresponding to a valid
          discriminant

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
            sage: G = HeckeTriangleGroup(n=4)
            sage: sorted(G.rational_period_functions(k=4, D=12))
            [(z^4 - 1)/z^4]
            sage: sorted(G.rational_period_functions(k=-2, D=12))
            [-z^2 + 1, 4*lam*z^2 - 4*lam]
            sage: sorted(G.rational_period_functions(k=2, D=14))
            [(24*z^6 - 120*z^4 + 120*z^2 - 24)/(9*z^8 - 80*z^6 + 146*z^4 - 80*z^2 + 9),
             (24*z^6 - 120*z^4 + 120*z^2 - 24)/(9*z^8 - 80*z^6 + 146*z^4 - 80*z^2 + 9),
             1/z,
             (z^2 - 1)/z^2]
            sage: sorted(G.rational_period_functions(k=-4, D=14))
            [-16*z^4 + 16, -z^4 + 1, 16*z^4 - 16]
        """
        try:
            k = ZZ(k)
            if not ZZ(2).divides(k):
                raise TypeError
        except TypeError:
            raise ValueError(f"k={k} has to be an even integer!")

        z = PolynomialRing(self.base_ring(), 'z').gen()

        R = []
        if k != 0:
            R.append(ZZ(1) - z**(-k))
        if k == 2:
            R.append(z**(-1))

        L = self.class_representatives(D=D, primitive=True)
        for v in L:
            rat = v.rational_period_function(k)
            if rat != 0:
                R.append(rat)

        return R



r"""
Common parametrized surfaces in 3D.

AUTHORS::

- Joris Vankerschaver (2012-06-16)
"""
#*****************************************************************************
#       Copyright (C) 2010  Joris Vankerschaver <joris.vankerschaver@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************


from sage.symbolic.constants import pi
from sage.functions.log import log
from sage.functions.trig import sin, cos, tan
from sage.functions.hyperbolic import cosh, tanh
from sage.symbolic.ring import var
from sage.geometry.riemannian_manifolds.parametrized_surface3d import \
    ParametrizedSurface3D


class SurfaceGenerators:
    """
    A class consisting of generators for several common parametrized surfaces
    in 3D.
    """
    @staticmethod
    def Catenoid(c=1, name='Catenoid'):
        r"""
        Return a catenoid surface, with parametric representation.

        .. MATH::

            \begin{aligned}
              x(u, v) & = c \cosh(v/c) \cos(u); \\
              y(u, v) & = c \cosh(v/c) \sin(u); \\
              z(u, v) & = v.
            \end{aligned}

        INPUT:

        - ``c`` -- surface parameter

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Catenoid`.

        EXAMPLES::

            sage: cat = surfaces.Catenoid(); cat
            Parametrized surface ('Catenoid') with equation (cos(u)*cosh(v), cosh(v)*sin(u), v)
            sage: cat.plot()                                                            # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        catenoid_eq = [c*cosh(v/c)*cos(u), c*cosh(v/c)*sin(u), v]
        coords = ((u, 0, 2 * pi), (v, -1, 1))
        return ParametrizedSurface3D(catenoid_eq, coords, name)

    @staticmethod
    def Crosscap(r=1, name='Crosscap'):
        r"""
        Return a crosscap surface, with parametrization.

        .. MATH::

            \begin{aligned}
              x(u, v) & = r(1 + \cos(v)) \cos(u); \\
              y(u, v) & = r(1 + \cos(v)) \sin(u); \\
              z(u, v) & = - r\tanh(u - \pi) \sin(v).
            \end{aligned}

        INPUT:

        - ``r`` -- surface parameter

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Cross-cap`.

        EXAMPLES::

            sage: crosscap = surfaces.Crosscap(); crosscap
            Parametrized surface ('Crosscap') with equation ((cos(v) + 1)*cos(u), (cos(v) + 1)*sin(u), -sin(v)*tanh(-pi + u))
            sage: crosscap.plot()                                                       # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        crosscap_eq = [r*(1+cos(v))*cos(u), r*(1+cos(v))*sin(u),
                       -tanh(u-pi)*r*sin(v)]
        coords = ((u, 0, 2 * pi), (v, 0, 2 * pi))
        return ParametrizedSurface3D(crosscap_eq, coords, name)

    @staticmethod
    def Dini(a=1, b=1, name="Dini's surface"):
        r"""
        Return Dini's surface, with parametrization.

        .. MATH::

            \begin{aligned}
              x(u, v) & = a \cos(u)\sin(v); \\
              y(u, v) & = a \sin(u)\sin(v); \\
              z(u, v) & = u + \log(\tan(v/2)) + \cos(v).
            \end{aligned}

        INPUT:

        - ``a``, ``b`` -- surface parameters

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Dini%27s_surface`.

        EXAMPLES::

            sage: dini = surfaces.Dini(a=3, b=4); dini
            Parametrized surface ('Dini's surface') with equation (3*cos(u)*sin(v), 3*sin(u)*sin(v), 4*u + 3*cos(v) + 3*log(tan(1/2*v)))
            sage: dini.plot()                                                           # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        dini_eq = [a*cos(u)*sin(v), a*sin(u)*sin(v),
                   a*(cos(v) + log(tan(v/2))) + b*u]
        coords = ((u, 0, 2 * pi), (v, 0, 2 * pi))
        return ParametrizedSurface3D(dini_eq, coords, name)

    @staticmethod
    def Ellipsoid(center=(0, 0, 0), axes=(1, 1, 1), name='Ellipsoid'):
        r"""
        Return an ellipsoid centered at ``center`` whose semi-principal axes
        have lengths given by the components of ``axes``. The
        parametrization of the ellipsoid is given by

        .. MATH::

            \begin{aligned}
              x(u, v) & = x_0 + a \cos(u) \cos(v); \\
              y(u, v) & = y_0 + b \sin(u) \cos(v); \\
              z(u, v) & = z_0 + c \sin(v).
            \end{aligned}

        INPUT:

        - ``center`` -- 3-tuple; coordinates of the center of the ellipsoid

        - ``axes`` -- 3-tuple; lengths of the semi-principal axes

        - ``name`` -- string; name of the ellipsoid

        For more information, see :wikipedia:`Ellipsoid`.

        EXAMPLES::

            sage: ell = surfaces.Ellipsoid(axes=(1, 2, 3)); ell
            Parametrized surface ('Ellipsoid') with equation (cos(u)*cos(v), 2*cos(v)*sin(u), 3*sin(v))
            sage: ell.plot()                                                            # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        x, y, z = center
        a, b, c = axes
        ellipsoid_parametric_eq = [x + a*cos(u)*cos(v),
                                   y + b*sin(u)*cos(v),
                                   z + c*sin(v)]
        coords = ((u, 0, 2 * pi), (v, -pi / 2, pi / 2))
        return ParametrizedSurface3D(ellipsoid_parametric_eq, coords, name)

    @staticmethod
    def Enneper(name="Enneper's surface"):
        r"""
        Return Enneper's surface, with parametrization.

        .. MATH::

            \begin{aligned}
              x(u, v) & = u(1 - u^2/3 + v^2)/3; \\
              y(u, v) & = -v(1 - v^2/3 + u^2)/3; \\
              z(u, v) & = (u^2 - v^2)/3.
            \end{aligned}

        INPUT:

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Enneper_surface`.

        EXAMPLES::

            sage: enn = surfaces.Enneper(); enn
            Parametrized surface ('Enneper's surface') with equation (-1/9*(u^2 - 3*v^2 - 3)*u, -1/9*(3*u^2 - v^2 + 3)*v, 1/3*u^2 - 1/3*v^2)
            sage: enn.plot()                                                            # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        enneper_eq = [u*(1-u**2/3+v**2)/3, -v*(1-v**2/3+u**2)/3, (u**2-v**2)/3]
        coords = ((u, -3, 3), (v, -3, 3))
        return ParametrizedSurface3D(enneper_eq, coords, name)

    @staticmethod
    def Helicoid(h=1, name='Helicoid'):
        r"""
        Return a helicoid surface, with parametrization.

        .. MATH::

            \begin{aligned}
              x(\rho, \theta) & = \rho \cos(\theta); \\
              y(\rho, \theta) & = \rho \sin(\theta); \\
              z(\rho, \theta) & = h\theta/(2\pi).
            \end{aligned}

        INPUT:

        - ``h`` -- distance along the z-axis between two
          successive turns of the helicoid

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Helicoid`.

        EXAMPLES::

            sage: helicoid = surfaces.Helicoid(h=2); helicoid
            Parametrized surface ('Helicoid') with equation (rho*cos(theta), rho*sin(theta), theta/pi)
            sage: helicoid.plot()                                                       # needs sage.plot
            Graphics3d Object
        """
        rho, theta = var('rho, theta')
        helicoid_eq = [rho*cos(theta), rho*sin(theta), h*theta/(2*pi)]
        coords = ((rho, -2, 2), (theta, 0, 2 * pi))
        return ParametrizedSurface3D(helicoid_eq, coords, name)

    @staticmethod
    def Klein(r=1, name="Klein bottle"):
        r"""
        Return the Klein bottle, in the figure-8 parametrization given by

        .. MATH::

            \begin{aligned}
              x(u, v) & = (r + \cos(u/2)\cos(v) - \sin(u/2)\sin(2v)) \cos(u); \\
              y(u, v) & = (r + \cos(u/2)\cos(v) - \sin(u/2)\sin(2v)) \sin(u); \\
              z(u, v) & = \sin(u/2)\cos(v) + \cos(u/2)\sin(2v).
            \end{aligned}

        INPUT:

        - ``r`` -- radius of the "figure-8" circle

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Klein_bottle`.

        EXAMPLES::

            sage: klein = surfaces.Klein(); klein
            Parametrized surface ('Klein bottle') with equation (-(sin(1/2*u)*sin(2*v) - cos(1/2*u)*sin(v) - 1)*cos(u), -(sin(1/2*u)*sin(2*v) - cos(1/2*u)*sin(v) - 1)*sin(u), cos(1/2*u)*sin(2*v) + sin(1/2*u)*sin(v))
            sage: klein.plot()                                                          # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        x = (r + cos(u/2)*sin(v) - sin(u/2)*sin(2*v))*cos(u)
        y = (r + cos(u/2)*sin(v) - sin(u/2)*sin(2*v))*sin(u)
        z = sin(u/2)*sin(v) + cos(u/2)*sin(2*v)
        klein_eq = [x, y, z]
        coords = ((u, 0, 2 * pi), (v, 0, 2 * pi))

        return ParametrizedSurface3D(klein_eq, coords, name)

    @staticmethod
    def MonkeySaddle(name="Monkey saddle"):
        r"""
        Return a monkey saddle surface, with equation

        .. MATH::

            z = x^3 - 3xy^2.

        INPUT:

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Monkey_saddle`.

        EXAMPLES::

            sage: saddle = surfaces.MonkeySaddle(); saddle
            Parametrized surface ('Monkey saddle') with equation (u, v, u^3 - 3*u*v^2)
            sage: saddle.plot()                                                         # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        monkey_eq = [u, v, u**3 - 3*u*v**2]
        coords = ((u, -2, 2), (v, -2, 2))

        return ParametrizedSurface3D(monkey_eq, coords, name)

    @staticmethod
    def Paraboloid(a=1, b=1, c=1, elliptic=True, name=None):
        r"""
        Return a paraboloid with equation.

        .. MATH::

            \frac{z}{c} = \pm \frac{x^2}{a^2} + \frac{y^2}{b^2}

        When the plus sign is selected, the paraboloid is elliptic. Otherwise
        the surface is a hyperbolic paraboloid.

        INPUT:

        - ``a``, ``b``, ``c`` -- surface parameters

        - ``elliptic`` -- boolean (default: ``True``); whether to create an
          elliptic or hyperbolic paraboloid

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Paraboloid`.

        EXAMPLES::

            sage: epar = surfaces.Paraboloid(1, 3, 2); epar
            Parametrized surface ('Elliptic paraboloid') with equation (u, v, 2*u^2 + 2/9*v^2)
            sage: epar.plot()                                                           # needs sage.plot
            Graphics3d Object

            sage: hpar = surfaces.Paraboloid(2, 3, 1, elliptic=False); hpar
            Parametrized surface ('Hyperbolic paraboloid') with equation (u, v, -1/4*u^2 + 1/9*v^2)
            sage: hpar.plot()                                                           # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        x = u
        y = v
        if elliptic:
            z = c*(v**2/b**2 + u**2/a**2)
        else:
            z = c*(v**2/b**2 - u**2/a**2)
        paraboloid_eq = [x, y, z]
        coords = ((u, -3, 3), (v, -3, 3))

        if name is None:
            if elliptic:
                name = "Elliptic paraboloid"
            else:
                name = "Hyperbolic paraboloid"

        return ParametrizedSurface3D(paraboloid_eq, coords, name)

    @staticmethod
    def Sphere(center=(0, 0, 0), R=1, name='Sphere'):
        r"""
        Return a sphere of radius ``R`` centered at ``center``.

        INPUT:

        - ``center`` -- 3-tuple; center of the sphere

        - ``R`` -- radius of the sphere

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Sphere`.

        EXAMPLES::

            sage: sphere = surfaces.Sphere(center=(0, 1, -1), R=2); sphere
            Parametrized surface ('Sphere') with equation (2*cos(u)*cos(v), 2*cos(v)*sin(u) + 1, 2*sin(v) - 1)
            sage: sphere.plot()                                                         # needs sage.plot
            Graphics3d Object

        Note that the radius of the sphere can be negative. The surface thus
        obtained is equal to the sphere (or part thereof) with positive radius,
        whose coordinate functions have been multiplied by -1. Compare for
        instant the first octant of the unit sphere with positive radius::

            sage: octant1 = surfaces.Sphere(R=1); octant1
            Parametrized surface ('Sphere') with equation (cos(u)*cos(v), cos(v)*sin(u), sin(v))
            sage: octant1.plot((0, pi/2), (0, pi/2))                                    # needs sage.plot
            Graphics3d Object

        with the first octant of the unit sphere with negative radius::

            sage: octant2 = surfaces.Sphere(R=-1); octant2
            Parametrized surface ('Sphere') with equation (-cos(u)*cos(v), -cos(v)*sin(u), -sin(v))
            sage: octant2.plot((0, pi/2), (0, pi/2))                                    # needs sage.plot
            Graphics3d Object
        """
        return SurfaceGenerators.Ellipsoid(center, (R, R, R), name)

    @staticmethod
    def Torus(r=2, R=3, name='Torus'):
        r"""
        Return a torus obtained by revolving a circle of radius ``r`` around
        a coplanar axis ``R`` units away from the center of the circle. The
        parametrization used is

        .. MATH::

            \begin{aligned}
              x(u, v) & = (R + r \cos(v)) \cos(u); \\
              y(u, v) & = (R + r \cos(v)) \sin(u); \\
              z(u, v) & = r \sin(v).
            \end{aligned}

        INPUT:

        - ``r``, ``R`` -- minor and major radius of the torus

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Torus`.

        EXAMPLES::

            sage: torus = surfaces.Torus(); torus
            Parametrized surface ('Torus') with equation ((2*cos(v) + 3)*cos(u), (2*cos(v) + 3)*sin(u), 2*sin(v))
            sage: torus.plot()                                                          # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        torus_eq = [(R+r*cos(v))*cos(u), (R+r*cos(v))*sin(u), r*sin(v)]
        coords = ((u, 0, 2 * pi), (v, 0, 2 * pi))
        return ParametrizedSurface3D(torus_eq, coords, name)

    @staticmethod
    def WhitneyUmbrella(name="Whitney's umbrella"):
        r"""
        Return Whitney's umbrella, with parametric representation

        .. MATH::

            x(u, v) = uv, \quad y(u, v) = u, \quad z(u, v) = v^2.

        INPUT:

        - ``name`` -- string; name of the surface

        For more information, see :wikipedia:`Whitney_umbrella`.

        EXAMPLES::

            sage: whitney = surfaces.WhitneyUmbrella(); whitney
            Parametrized surface ('Whitney's umbrella') with equation (u*v, u, v^2)
            sage: whitney.plot()                                                        # needs sage.plot
            Graphics3d Object
        """
        u, v = var('u, v')
        whitney_eq = [u*v, u, v**2]
        coords = ((u, -1, 1), (v, -1, 1))
        return ParametrizedSurface3D(whitney_eq, coords, name)


# Easy access to the surface generators
surfaces = SurfaceGenerators()



r"""
Functions for computing the profiles of sub-Hopf algebras of the
mod p Steenrod algebra.

In particular, the main functions are :func:`profile_elt` and
:func:`enveloping_profile_elements`, which compute the profile
function for the smallest sub-Hopf algebra of the Steenrod algebra
containing the given elements.

See :func:`SteenrodAlgebra
<sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra>`
:mod:`sage.algebras.steenrod.steenrod_algebra` for information about
profile functions.

AUTHORS:

- Robert R. Bruner, Michael J. Catanzaro (2012): initial version
- John Palmieri (2022): cleanup, modifications
"""

# ****************************************************************************
#       Copyright (C) 2011 Robert R. Bruner <rrb@math.wayne.edu> and
#                          Michael J. Catanzaro <mike@math.wayne.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from itertools import zip_longest
from sage.rings.integer_ring import ZZ


def profile_elt(elt, char=2):
    """
    Return the smallest sub-Hopf algebra containing ``elt``.

    INPUT:

    - ``elt`` -- element of the Steenrod algebra (or a sub-Hopf algebra
      of it) or list(s) representing it
    - ``char`` -- (default: 2) the characteristic

    ``elt`` could also be a list (when ``char=2``) or a pair of lists
    (otherwise), in which case it is treated as corresponding to an
    element of the Steenrod algebra: ``(a, b, c) <-> Sq(a, b, c)`` or
    ``((a, b, c), (x, y, z)) <-> Q_a Q_b Q_c P(x, y, z)``.

    OUTPUT: the profile function corresponding to the smallest
    sub-Hopf algebra containing the element passed

    EXAMPLES::

        sage: from sage.modules.fp_graded.steenrod.profile import profile_elt
        sage: A2 = SteenrodAlgebra(2)
        sage: profile_elt(A2.Sq(2))
        (2, 1)
        sage: profile_elt(A2.Sq(4,8))
        (3, 4, 3, 2, 1)

        sage: B = SteenrodAlgebra(3)
        sage: b = B.an_element(); b
        2 Q_1 Q_3 P(2,1)
        sage: profile_elt(b, char=3)
        ((1, 1), (1, 2, 1, 2))
        sage: profile_elt(B.P(2,1), char=3)
        ((1, 1), ())
        sage: profile_elt(B.Q(2), char=3)
        ((0,), (1, 1, 2))
    """
    try:
        # Convert element of the Steenrod algebra to iterable(s).
        elt = elt.leading_support()
    except AttributeError:
        # Not in the Steenrod algebra so presumably already in the
        # right form.
        pass

    if char == 2:
        minprofile = [max(0, ZZ(n).exact_log(char) + 1) for n in elt]
        return find_min_profile(minprofile, char)

    # odd primes:
    alistQ, alistP = elt
    minprofileP = [max(0, ZZ(n).exact_log(char) + 1) for n in alistP]
    if not alistQ:
        minpQ = []
    else:
        minpQ = [1] * (max(alistQ) + 1)
        for j in alistQ:
            minpQ[j] = 2
    return find_min_profile((minprofileP, minpQ), char=char)


def enveloping_profile_elements(alist, char=2):
    r"""
    Return the profile function for the smallest sub-Hopf algebra
    containing the list of elements passed.

    INPUT:

    - ``alist`` -- list of Steenrod algebra elements
    - ``char`` -- (default: 2) the characteristic

    As with :func:`profile_elt`, the entries of ``alist`` could also
    be iterables or pairs of iterables.

    OUTPUT: the profile function for the minimum sub-algebra
    containing all the elements of ``alist``

    EXAMPLES::

        sage: from sage.modules.fp_graded.steenrod.profile import enveloping_profile_elements
        sage: enveloping_profile_elements([Sq(2),Sq(4)])
        (3, 2, 1)
        sage: enveloping_profile_elements([Sq(4)])
        (3, 2, 1)
        sage: enveloping_profile_elements([Sq(2,1,2),Sq(7)])
        (3, 2, 2, 1)

        sage: B = SteenrodAlgebra(3)
        sage: enveloping_profile_elements([B.P(2,1), B.P(0,0,3)], char=3)
        ((1, 1, 2, 1), ())
        sage: enveloping_profile_elements([B.P(3,1)], char=3)
        ((2, 1), ())
        sage: enveloping_profile_elements([B.P(2,1), B.P(0,0,3), B.Q(2)], char=3)
        ((1, 1, 2, 1), (1, 1, 2))
    """
    if char == 2:
        profiles = [profile_elt(x) for x in alist if x != 0]
        if not profiles:
            return (0,)
        # zip_longest doesn't do the right thing with lists of length
        # 1, but this case is simple:
        if len(profiles) == 1:
            return profiles[0]
        return find_min_profile(max(*a) for a in zip_longest(*profiles, fillvalue=0))

    # odd primes:
    profiles = [profile_elt(x, char) for x in alist if x != 0]
    if len(profiles) == 1:
        return profiles[0]
    profiles_P = [x[0] for x in profiles]
    profiles_Q = [x[1] for x in profiles]
    if not profiles_P and not profiles_Q:
        return ((0,), (0,))
    maxP = [max(*a) for a in zip_longest(*profiles_P, fillvalue=0)]
    maxQ = [max(*a) for a in zip_longest(*profiles_Q, fillvalue=0)]
    return find_min_profile([maxP, maxQ], char=char)


def find_min_profile(prof, char=2):
    r"""
    Return the smallest valid profile function containing a tuple of
    nonnegative integers.

    INPUT:

    - ``prof`` -- list or tuple of nonnegative integers
    - ``char`` -- (default: 2) the characteristic

    OUTPUT: a valid profile containing ``prof``

    A profile function `e` must satisfy `e(r) \geq \min( e(r-i) - i,
    e(i))` for all `0 < i < r`, and at odd primes, if `k(i+j) = 1`,
    then either `e(i) \leq j` or `k(j) = 1` for all `i \geq 1`, `j
    \geq 0`. We use these inequalities to generate the smallest
    profile function `e` satisfying `e(r) \geq prof(r)` for each `r`
    when `char=2`, and similarly at odd primes.

    EXAMPLES::

        sage: from sage.modules.fp_graded.steenrod.profile import find_min_profile
        sage: find_min_profile([1,2])
        (1, 2, 1)
        sage: find_min_profile([2,1])
        (2, 1)
        sage: find_min_profile([1,2,3])
        (1, 2, 3, 1, 1)
        sage: find_min_profile([4])
        (4, 3, 2, 1)

        sage: find_min_profile([[4], []], char=3)
        ((4, 3, 2, 1), ())
        sage: find_min_profile([[1], [2]], char=3)
        ((1,), (2, 2))
        sage: find_min_profile([[], [2,1,1,2]], char=3)
        ((0,), (2, 1, 1, 2))
    """
    if char == 2:
        if not prof:
            return (0,)
        # Add a zero to the front so that the relevant part of the
        # pseudo-profile new is indexed starting with 1.
        new = [0] + list(prof)
        # new is probably too short. Increase its length by using the
        # defining inequality for a profile function:
        #    e(s+t) \geq min(e(s)-t, e(t)).
        # So if min(e(s)-t, e(t)) > 0, then the length must be at
        # least s+t.
        pad = 0
        new += [0] * max(new)
        for t in range(len(new)):
            for s in range(len(new)):
                if min(new[s] - t, new[t]) > 0:
                    pad = max(pad, s + t)
        e = [0] * len(new)
        # Now compute the new profile e.
        for r in range(len(e)):
            m = max((min(e[r - i] - i, e[i]) for i in range(1, r)), default=0)
            e[r] = max(m, new[r])
        # Strip trailing zeroes.
        while e[-1] == 0:
            e = e[:-1]
        return tuple(e[1:])

    # odd primes:
    pP = list(prof[0])
    pQ = list(prof[1])
    P = find_min_profile(pP, char=2)
    if not pQ:
        return (P, tuple(pQ))
    # newQ: dictionary of the form {index: value} where value is
    # either 1 or 2.
    maxP = max(P)
    newQ = list(pQ) + [None] * maxP
    for j in range(len(pQ)):
        if newQ[j] == 2:
            for i in range(maxP):
                if P[i] > j:
                    newQ[i + 1 + j] = 2
    # Strip all of the None values
    # Do it from the back to minimize reshuffles and keep the index matching
    for i in range(len(newQ) - 1, len(pQ) - 1, -1):
        if newQ[i] is None:
            del newQ[i]
    return (P, tuple(newQ))



r"""
Catalog of discrete dynamical systems

This module contains constructors for several specific discrete
dynamical systems.
These are accessible through
:mod:`sage.dynamics.finite_dynamical_system_catalog. <sage.dynamics.finite_dynamical_system_catalog>`
or just through ``finite_dynamical_systems.``
(type either of these in Sage and hit ``tab`` for a list).

AUTHORS:

- Darij Grinberg, Tom Roby (2018): initial version

Functions
=========
"""
#*****************************************************************************
#       Copyright (C) 2018 Darij Grinberg <darijgrinberg@gmail.com>,
#                     2018 Tom Roby <tomrobyuconn@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.dynamics.finite_dynamical_system import DiscreteDynamicalSystem, \
        FiniteDynamicalSystem, InvertibleDiscreteDynamicalSystem, \
        InvertibleFiniteDynamicalSystem


def permutation(pi, invertible=True):
    r"""
    Return the invertible finite discrete dynamical system
    induced by the permutation ``pi`` of the set
    `\{1, 2, \ldots, n\}`.

    The permutation ``pi`` can be provided as an instance
    of :class:`Permutation`, but also as anything that can
    be cast into that class.

    See also :meth:`one_line` for a more general
    construction, which takes any map from
    `\{1, 2, \ldots, n\}` to `\{1, 2, \ldots, n\}` given in
    one-line notation, and builds a (not necessarily
    invertible) discrete dynamical system out of it.

    EXAMPLES::

        sage: F = finite_dynamical_systems.permutation([3, 5, 4, 1, 2])
        sage: F.verify_inverse_evolution()
        True
        sage: sorted(F.orbit_lengths())
        [2, 3]
        sage: F.orbit(3)
        [3, 4, 1]
        sage: F.is_homomesic(lambda x: 1)
        True
        sage: F.is_homomesic(lambda x: x)
        False
    """
    from sage.combinat.permutation import Permutation
    pi = Permutation(pi)
    n = len(pi)
    X = range(1, n+1)
    return InvertibleFiniteDynamicalSystem(X, pi, inverse=pi.inverse(), create_tuple=True)


def one_line(xs):
    r"""
    Return the finite discrete dynamical system
    with ground set `\{1, 2, \ldots, n\}` and evolution
    sending each `i` to `x_i`,
    where `(x_1, x_2, \ldots, x_n)` is the argument ``xs``
    provided.

    EXAMPLES::

        sage: F = finite_dynamical_systems.one_line([2, 2, 1, 2, 3])
        sage: F.orbit(3)
        [3, 1, 2]
        sage: F.orbit(5)
        [5, 3, 1, 2]
        sage: F.orbit(2)
        [2]
    """
    n = len(xs)
    X = range(1, n + 1)
    xs2 = tuple(xs)

    def pi(i):
        return xs2[i - 1]
    return FiniteDynamicalSystem(X, pi, create_tuple=True)


def bitstring_rotation(n, ones=None):
    r"""
    Return the invertible finite discrete dynamical system
    consisting of all bitstrings of size `n` (that is, of
    all `n`-tuples `(i_1, \ldots, i_n) \in \{0, 1\}^n`),
    evolving by cyclic rotation.

    If the optional parameter ``ones`` is provided, the
    system is restricted only to those bitstrings whose
    number of ones is the value of this parameter.

    EXAMPLES::

        sage: F = finite_dynamical_systems.bitstring_rotation(5)
        sage: sorted(F.orbit_lengths())
        [1, 1, 5, 5, 5, 5, 5, 5]
        sage: F.orbit((0, 1, 1, 0, 1))
        [(0, 1, 1, 0, 1),
         (1, 1, 0, 1, 0),
         (1, 0, 1, 0, 1),
         (0, 1, 0, 1, 1),
         (1, 0, 1, 1, 0)]
        sage: F.is_homomesic(lambda x: sum(1 for i in range(5) for j in range(i) if x[j] > x[i]))
        False
        sage: F.is_homomesic(lambda x: x[0])
        False
        sage: F = finite_dynamical_systems.bitstring_rotation(5, ones=3)
        sage: F.verify_inverse_evolution()
        True
        sage: sorted(F.orbit_lengths())
        [5, 5]
        sage: F.orbit((0, 1, 1, 0, 1))
        [(0, 1, 1, 0, 1),
         (1, 1, 0, 1, 0),
         (1, 0, 1, 0, 1),
         (0, 1, 0, 1, 1),
         (1, 0, 1, 1, 0)]
        sage: F.is_homomesic(lambda x: sum(1 for i in range(5) for j in range(i) if x[j] > x[i]))
        True
        sage: F.is_homomesic(lambda x: x[0])
        True
    """
    if ones is None:
        from sage.categories.cartesian_product import cartesian_product
        X = cartesian_product([[0,1]] * n)
    else:
        from itertools import combinations
        X = [tuple((1 if i in cs else 0) for i in range(n))
             for cs in combinations(range(n), ones)]
    if n == 0:
        phi = lambda x: x
        psi = phi
    else:
        phi = lambda x: x[1:] + (x[0],)
        psi = lambda x: (x[-1],) + x[:-1]
    return InvertibleFiniteDynamicalSystem(X, phi, inverse=psi)


def striker_sweep(E, predicate, elements, lazy=False):
    r"""
    Return the invertible finite discrete dynamical system
    on all subsets of a finite set ``E`` satisfying a
    boolean predicate ``predicate``, where evolution is
    the "Striker sweep" -- i.e., the composition of
    Striker toggles corresponding to the elements in the
    iterable ``elements`` (from first to last).

    Let `E` be a finite set.
    Let `\mathcal{L}` be a subset of the powerset of `E`.
    (In this implementation, `\mathcal{L}` should be
    specified via the boolean predicate ``predicate``,
    which takes a subset `F` of `E` and returns the truth
    value of `F \in \mathcal{L}`.)
    For any `e \in E`, the *Striker toggle* `t_e` is
    the involution of the set `\mathcal{L}` that sends
    each `F \in \mathcal{L}` to the symmetric difference
    `F \triangle \{ e \}` if this symmetric difference
    is in `\mathcal{L}`, and otherwise to `F` itself.
    If `(e_1, e_2, \ldots, e_k)` is a finite sequence of
    elements of `E` (to be provided as a tuple, via the
    argument ``elements``), then the *Striker sweep*
    corresponding to this sequence is the composition of
    maps
    `t_{e_k} \circ t_{e_{k-1}} \circ \cdots \circ t_{e_1}`.

    This generalizes classical constructions such as
    rowmotion on order ideals.

    The optional argument ``lazy`` can be set to
    ``True``; in that case, the ground set of the
    dynamical system will not be explicitly computed.

    .. WARNING::

        The sequence ``elements`` should be provided as
        a tuple, or as a list that is guaranteed not to
        be mutated (as otherwise, mutation will corrupt
        this DDS).

    .. TODO::

        Implement the ``lazy=True`` case.
        This should differ in that ``X`` is no longer a
        list, but an enumerated set.
        But how to build an enumerated set filtering a
        given enumerated set according to a predicate?

    EXAMPLES::

        sage: StS = finite_dynamical_systems.striker_sweep
        sage: E = range(1, 5)
        sage: lac = lambda S: all(s + 1 not in S for s in S) # lacunarity predicate
        sage: F = StS(E, lac, [1, 2, 3, 4])
        sage: F.ground_set()
        [{}, {1}, {2}, {3}, {4}, {1, 3}, {1, 4}, {2, 4}]
        sage: F.evolution()(Set([2, 4]))
        {}
        sage: F.evolution()(Set([]))
        {1, 3}
        sage: F.evolution()(Set([1, 3]))
        {4}
        sage: F.evolution()(Set([4]))
        {1}
        sage: F.inverse_evolution()(Set([1]))
        {4}
        sage: F.verify_inverse_evolution()
        True
        sage: sorted(F.orbit_lengths())
        [3, 5]
        sage: F.orbit(Set([2, 4]))
        [{2, 4}, {}, {1, 3}, {4}, {1}]
        sage: F.is_homomesic(lambda S: S.cardinality())
        False
        sage: F.is_homomesic(lambda S: bool(1 in S) - bool(4 in S), find_average=True)
        0
        sage: F.is_homomesic(lambda S: bool(2 in S) - bool(3 in S), find_average=True)
        0
        sage: F.is_homomesic(lambda S: bool(1 in S), find_average=True)
        False
    """
    from sage.combinat.subset import Subsets
    from sage.sets.set import Set
    X = [F for F in Subsets(E) if predicate(F)]

    def phi(F):
        for e in elements:
            G = F.symmetric_difference(Set([e]))
            if predicate(G):
                F = G
        return F

    def psi(F):
        for e in reversed(elements):
            G = F.symmetric_difference(Set([e]))
            if predicate(G):
                F = G
        return F
    return InvertibleFiniteDynamicalSystem(X, phi, inverse=psi)


def syt_promotion(lam):
    r"""
    Return the invertible finite discrete dynamical system
    consisting of all standard tableaux of shape ``lam`` (a
    given partition) and evolving according to promotion.

    EXAMPLES::

        sage: F = finite_dynamical_systems.syt_promotion([4, 4, 4])
        sage: sorted(F.orbit_lengths())
        [3, 3, 4, 4, 4, 6, 6, 6, 6, 12, 12, 12, 12, 12, 12,
         12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
         12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
         12, 12, 12, 12, 12]
        sage: G = finite_dynamical_systems.syt_promotion([4, 3, 1])
        sage: sorted(G.orbit_lengths())
        [16, 22, 32]
        sage: G.verify_inverse_evolution()
        True
    """
    from sage.combinat.partition import Partition
    from sage.combinat.tableau import StandardTableaux
    lam = Partition(lam)
    X = StandardTableaux(lam)
    return InvertibleFiniteDynamicalSystem(X, lambda T: T.promotion(),
                                           inverse=lambda T: T.promotion_inverse())


def order_ideal_rowmotion(P):
    r"""
    Return the invertible finite discrete dynamical system
    consisting of all order ideals of the poset ``P``,
    evolving according to rowmotion.

    EXAMPLES::

        sage: P = RootSystem(["A", 6]).root_poset()
        sage: F = finite_dynamical_systems.order_ideal_rowmotion(P)
        sage: sorted(F.orbit_lengths())
        [2, 7, 7, 7, 7, 7, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
         14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14]
        sage: F.is_homomesic(lambda I: len(I))
        False
        sage: F.is_homomesic(lambda I: sum((-1)**(P.rank(i)) for i in I))
        True

        sage: P = RootSystem(["A", 3]).root_poset()
        sage: F = finite_dynamical_systems.order_ideal_rowmotion(P)
        sage: F.verify_inverse_evolution()
        True
    """
    from sage.sets.set import Set
    X = [Set(P.order_ideal(A)) for A in P.antichains()]
    # Using P.order_ideals_lattice() instead causes intransparency issues:
    # sage can't always do P.rowmotion(I) when I is in P.order_ideals_lattice().
    # Bug in P.order_ideals_lattice() when P is facade?
    phi = P.rowmotion

    def psi(I):  # inverse of rowmotion
        result = I
        for i in P.linear_extension():
            result = P.order_ideal_toggle(result, i)
        return result
    return InvertibleFiniteDynamicalSystem(X, phi, inverse=psi)


def semidistributive_rowmotion(L):
    r"""
    Return the invertible finite discrete dynamical system
    consisting of the elements of the semidistributive lattice ``L``,
    evolving according to (semidistributive) rowmotion.

    EXAMPLES::

        sage: L = posets.TamariLattice(3)
        sage: row = finite_dynamical_systems.semidistributive_rowmotion(L)
        sage: all(L.rowmotion_semidistributive(a) == row.evolution()(a) for a in L)
        True
    """
    H = L._hasse_diagram
    d = L._element_to_vertex_dict
    meet_irr = [u for u in H if sum(1 for _ in H.upper_covers_iterator(u)) == 1]
    join_irr = [u for u in H if sum(1 for _ in H.lower_covers_iterator(u)) == 1]
    kappa_dual = {L._vertex_to_element(u): L._vertex_to_element(H.kappa_dual(u))
                  for u in meet_irr}
    row0 = {a: L.join(kappa_dual[e] for e in L.canonical_meetands(a))
            for a in L}

    kappa = {L._vertex_to_element(u): L._vertex_to_element(H.kappa(u))
             for u in join_irr}
    row1 = {a: L.meet(kappa[e] for e in L.canonical_joinands(a))
            for a in L}
    return InvertibleFiniteDynamicalSystem(L, lambda a: row0[a], inverse=lambda a: row1[a])


def bulgarian_solitaire(n):
    r"""
    Return the finite discrete dynamical system defined
    by Bulgarian solitaire on partitions of size `n`.

    Let `n` be a nonnegative integer.
    Let `P` be the set of all integer partitions of
    size `n`.
    Let `B : P \to P` be the map that sends each
    partition
    `\lambda = (\lambda_1, \lambda_2, \ldots, \lambda_k)`
    of `n` (with all the `\lambda_i` positive) to
    `(\lambda_1 - 1, \lambda_2 - 1, \ldots, \lambda_k - 1, k)`,
    where zero entries have been removed and the remaining
    entries sorted into decreasing order.
    (For example,
    `B(5, 4, 2, 2, 1, 1) = (6, 4, 3, 1, 1)`.)
    This method yields the finite DDS whose ground set
    is `P` and whose evolution is `B`.

    EXAMPLES::

        sage: BS = finite_dynamical_systems.bulgarian_solitaire
        sage: BS(3).evolution()(Partition([3]))
        [2, 1]
        sage: BS(3).evolution()(Partition([2, 1]))
        [2, 1]
        sage: BS(3).evolution()(Partition([1, 1, 1]))
        [3]
        sage: BS(4).evolution()(Partition([4]))
        [3, 1]
        sage: BS(4).orbit(Partition([4]))
        [[4], [3, 1], [2, 2], [2, 1, 1]]
        sage: BS(4).orbit(Partition([3, 1]))
        [[3, 1], [2, 2], [2, 1, 1]]
        sage: BS(7).orbit(Partition([6, 1]), preperiod=True)
        ([[6, 1], [5, 2], [4, 2, 1], [3, 3, 1], [3, 2, 2], [3, 2, 1, 1]], 2)
        sage: BS(6).is_homomesic(lambda lam: len(lam))
        True
        sage: BS(6).is_homomesic(lambda lam: lam[0])
        True
        sage: BS(6).is_homomesic(lambda lam: lam[-1])
        True
        sage: BS(8).is_homomesic(lambda lam: len(lam))
        True
        sage: BS(8).is_homomesic(lambda lam: lam[0])
        True
        sage: BS(8).is_homomesic(lambda lam: lam[-1])
        False
    """
    from sage.combinat.partition import Partitions, _Partitions
    X = Partitions(n)

    def phi(lam):
        mu = [p - 1 for p in lam if p > 0]
        nu = sorted(mu + [len(lam)], reverse=True)
        return _Partitions(nu)

    return FiniteDynamicalSystem(X, phi)



# sage.doctest: needs sage.libs.pari
"""
Automorphisms of Quadratic Forms
"""
# ****************************************************************************
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.misc.cachefunc import cached_method
from sage.matrix.constructor import Matrix
from sage.rings.integer_ring import ZZ

from sage.modules.free_module import FreeModule
from sage.modules.free_module_element import vector
from sage.arith.misc import GCD


@cached_method
def basis_of_short_vectors(self, show_lengths=False):
    r"""
    Return a basis for `\ZZ^n` made of vectors with minimal lengths `Q(v)`.

    OUTPUT: a tuple of vectors, and optionally a tuple of values for each vector

    This uses :pari:`qfminim`.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.basis_of_short_vectors()
        ((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1))
        sage: Q.basis_of_short_vectors(True)
        (((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1)), (1, 3, 5, 7))

    The returned vectors are immutable::

        sage: v = Q.basis_of_short_vectors()[0]
        sage: v
        (1, 0, 0, 0)
        sage: v[0] = 0
        Traceback (most recent call last):
        ...
        ValueError: vector is immutable; please change a copy instead (use copy())
    """
    # Set an upper bound for the number of vectors to consider
    Max_number_of_vectors = 10000

    # Generate a PARI matrix for the associated Hessian matrix
    M_pari = self.__pari__()

    # Run through all possible minimal lengths to find a spanning set of vectors
    n = self.dim()
    M1 = Matrix([[0]])
    vec_len = 0
    while M1.rank() < n:
        vec_len += 1
        pari_mat = M_pari.qfminim(vec_len, Max_number_of_vectors)[2]
        number_of_vecs = ZZ(pari_mat.matsize()[1])
        vector_list = []
        for i in range(number_of_vecs):
            new_vec = vector([ZZ(x) for x in list(pari_mat[i])])
            vector_list.append(new_vec)

        # Make a matrix from the short vectors
        if vector_list:
            M1 = Matrix(vector_list)

    # Organize these vectors by length (and also introduce their negatives)
    max_len = vec_len // 2
    vector_list_by_length = [[] for _ in range(max_len + 1)]
    for v in vector_list:
        l = self(v)
        vector_list_by_length[l].append(v)
        vector_list_by_length[l].append(vector([-x for x in v]))

    # Make a matrix from the column vectors (in order of ascending length).
    sorted_list = [v for i in range(len(vector_list_by_length))
                   for v in vector_list_by_length[i]]
    sorted_matrix = Matrix(sorted_list).transpose()

    # Determine a basis of vectors of minimal length
    pivots = sorted_matrix.pivots()
    basis = tuple(sorted_matrix.column(i) for i in pivots)
    for v in basis:
        v.set_immutable()

    # Return the appropriate result
    if show_lengths:
        pivot_lengths = tuple(self(v) for v in basis)
        return basis, pivot_lengths
    return basis


def short_vector_list_up_to_length(self, len_bound, up_to_sign_flag=False):
    """
    Return a list of lists of short vectors `v`, sorted by length, with
    `Q(v) <` ``len_bound``.

    INPUT:

    - ``len_bound`` -- bound for the length of the vectors

    - ``up_to_sign_flag`` -- boolean (default: ``False``); if set to ``True``,
      then only one of the vectors of the pair `[v, -v]` is listed

    OUTPUT:

    A list of lists of vectors such that entry `[i]` contains all
    vectors of length `i`.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.short_vector_list_up_to_length(3)
        [[(0, 0, 0, 0)], [(1, 0, 0, 0), (-1, 0, 0, 0)], []]
        sage: Q.short_vector_list_up_to_length(4)
        [[(0, 0, 0, 0)],
         [(1, 0, 0, 0), (-1, 0, 0, 0)],
         [],
         [(0, 1, 0, 0), (0, -1, 0, 0)]]
        sage: Q.short_vector_list_up_to_length(5)
        [[(0, 0, 0, 0)],
         [(1, 0, 0, 0), (-1, 0, 0, 0)],
         [],
         [(0, 1, 0, 0), (0, -1, 0, 0)],
         [(1, 1, 0, 0),
          (-1, -1, 0, 0),
          (1, -1, 0, 0),
          (-1, 1, 0, 0),
          (2, 0, 0, 0),
          (-2, 0, 0, 0)]]
        sage: Q.short_vector_list_up_to_length(5, True)
        [[(0, 0, 0, 0)],
         [(1, 0, 0, 0)],
         [],
         [(0, 1, 0, 0)],
         [(1, 1, 0, 0), (1, -1, 0, 0), (2, 0, 0, 0)]]
        sage: m6 = matrix(6, [2, 1, 1, 1, -1, -1, 1, 2, 1, 1, -1, -1,
        ....:                 1, 1, 2, 0, -1, -1, 1, 1, 0, 2, 0, -1,
        ....:                 -1, -1, -1, 0, 2, 1, -1, -1, -1, -1, 1, 2])
        sage: Q = QuadraticForm(m6)
        sage: vs = Q.short_vector_list_up_to_length(8)
        sage: [len(vs[i]) for i in range(len(vs))]
        [1, 72, 270, 720, 936, 2160, 2214, 3600]

    The cases of ``len_bound < 2`` led to exception or infinite runtime before.

    ::

        sage: Q.short_vector_list_up_to_length(-1)
        []
        sage: Q.short_vector_list_up_to_length(0)
        []
        sage: Q.short_vector_list_up_to_length(1)
        [[(0, 0, 0, 0, 0, 0)]]

    In the case of quadratic forms that are not positive definite an error is raised.

    ::

        sage: QuadraticForm(matrix(2, [2, 0, 0, -2])).short_vector_list_up_to_length(3)
        Traceback (most recent call last):
        ...
        ValueError: Quadratic form must be positive definite in order to enumerate short vectors

    Check that PARI does not return vectors which are too long::

        sage: Q = QuadraticForm(matrix(2, [72, 12, 12, 120]))
        sage: len_bound_pari = 2*22953421 - 2; len_bound_pari
        45906840
        sage: vs = list(Q.__pari__().qfminim(len_bound_pari)[2])  # long time (18s on sage.math, 2014)
        sage: v = vs[0]; v  # long time
        [66, -623]~
        sage: v.Vec() * Q.__pari__() * v  # long time
        45902280
    """
    if not self.is_positive_definite():
        raise ValueError("Quadratic form must be positive definite "
                         "in order to enumerate short vectors")

    from sage.libs.pari import pari

    if len_bound <= 0:
        return []

    # Free module in which the vectors live
    V = FreeModule(ZZ, self.dim())

    # Adjust length for PARI. We need to subtract 1 because PARI returns
    # returns vectors of length less than or equal to b, but we want
    # strictly less. We need to double because the matrix is doubled.
    len_bound_pari = 2 * (len_bound - 1)

    # Call PARI's qfminim()
    parilist = self.__pari__().qfminim(len_bound_pari)[2].Vec()

    # List of lengths
    parilens = pari(r"(M,v) -> vector(#v, i, (v[i]~ * M * v[i])\2)")(self, parilist)

    # Sort the vectors into lists by their length
    vec_sorted_list = [[] for i in range(len_bound)]
    for i in range(len(parilist)):
        length = int(parilens[i])
        # In certain trivial cases, PARI can sometimes return longer
        # vectors than requested.
        if length < len_bound:
            sagevec = V(list(parilist[i]))
            vec_sorted_list[length].append(sagevec)
            if not up_to_sign_flag:
                vec_sorted_list[length].append(-sagevec)

    # Add the zero vector by hand
    vec_sorted_list[0].append(V.zero_vector())

    return vec_sorted_list


def short_primitive_vector_list_up_to_length(self, len_bound, up_to_sign_flag=False):
    r"""
    Return a list of lists of short primitive vectors `v`, sorted by length, with
    `Q(v) <` ``len_bound``.  The list in output `[i]` indexes all vectors of
    length `i`.  If the ``up_to_sign_flag`` is set to ``True``, then only one of
    the vectors of the pair `[v, -v]` is listed.

    .. NOTE::

        This processes the PARI/GP output to always give elements of type `\ZZ`.

    OUTPUT: list of lists of vectors

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.short_vector_list_up_to_length(5, True)
        [[(0, 0, 0, 0)],
         [(1, 0, 0, 0)],
         [],
         [(0, 1, 0, 0)],
         [(1, 1, 0, 0), (1, -1, 0, 0), (2, 0, 0, 0)]]
        sage: Q.short_primitive_vector_list_up_to_length(5, True)
        [[], [(1, 0, 0, 0)], [], [(0, 1, 0, 0)], [(1, 1, 0, 0), (1, -1, 0, 0)]]
    """
    # Get a list of short vectors
    full_vec_list = self.short_vector_list_up_to_length(len_bound, up_to_sign_flag)

    # Make a new list of the primitive vectors
    prim_vec_list = [[v for v in L if GCD(v) == 1]
                     for L in full_vec_list]

    # Return the list of primitive vectors
    return prim_vec_list


def _compute_automorphisms(self):
    """
    Call PARI to compute the automorphism group of the quadratic form.

    This uses :pari:`qfauto`.

    OUTPUT: none, this just caches the result

    TESTS::

        sage: DiagonalQuadraticForm(ZZ, [-1,1,1])._compute_automorphisms()
        Traceback (most recent call last):
        ...
        ValueError: not a definite form in QuadraticForm.automorphisms()
        sage: DiagonalQuadraticForm(GF(5), [1,1,1])._compute_automorphisms()
        Traceback (most recent call last):
        ...
        NotImplementedError: computing the automorphism group of a quadratic form is only supported over ZZ
    """
    if self.base_ring() is not ZZ:
        raise NotImplementedError("computing the automorphism group of a quadratic form is only supported over ZZ")
    if not self.is_definite():
        raise ValueError("not a definite form in QuadraticForm.automorphisms()")

    if hasattr(self, "__automorphisms_pari"):
        return

    A = self.__pari__().qfauto()
    self.__number_of_automorphisms = A[0]
    self.__automorphisms_pari = A[1]


def automorphism_group(self):
    """
    Return the group of automorphisms of the quadratic form.

    OUTPUT: a :class:`MatrixGroup`

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.automorphism_group()
        Matrix group over Rational Field with 3 generators (
        [ 0  0  1]  [1 0 0]  [ 1  0  0]
        [-1  0  0]  [0 0 1]  [ 0 -1  0]
        [ 0  1  0], [0 1 0], [ 0  0  1]
        )

    ::

        sage: DiagonalQuadraticForm(ZZ, [1,3,5,7]).automorphism_group()
        Matrix group over Rational Field with 4 generators (
        [-1  0  0  0]  [ 1  0  0  0]  [ 1  0  0  0]  [ 1  0  0  0]
        [ 0 -1  0  0]  [ 0 -1  0  0]  [ 0  1  0  0]  [ 0  1  0  0]
        [ 0  0 -1  0]  [ 0  0  1  0]  [ 0  0 -1  0]  [ 0  0  1  0]
        [ 0  0  0 -1], [ 0  0  0  1], [ 0  0  0  1], [ 0  0  0 -1]
        )

    The smallest possible automorphism group has order two, since we
    can always change all signs::

        sage: Q = QuadraticForm(ZZ, 3, [2, 1, 2, 2, 1, 3])
        sage: Q.automorphism_group()
        Matrix group over Rational Field with 1 generators (
        [-1  0  0]
        [ 0 -1  0]
        [ 0  0 -1]
        )
    """
    self._compute_automorphisms()

    from sage.matrix.matrix_space import MatrixSpace
    from sage.groups.matrix_gps.finitely_generated import MatrixGroup
    MS = MatrixSpace(self.base_ring().fraction_field(), self.dim(), self.dim())
    gens = [MS(x.sage()) for x in self.__automorphisms_pari]
    return MatrixGroup(gens)


def automorphisms(self):
    """
    Return the list of the automorphisms of the quadratic form.

    OUTPUT: list of matrices

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.number_of_automorphisms()
        48
        sage: 2^3 * factorial(3)
        48
        sage: len(Q.automorphisms())                                                    # needs sage.libs.gap
        48

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.number_of_automorphisms()
        16
        sage: aut = Q.automorphisms()                                                   # needs sage.libs.gap
        sage: len(aut)                                                                  # needs sage.libs.gap
        16
        sage: all(Q(M) == Q for M in aut)                                               # needs sage.libs.gap
        True

        sage: Q = QuadraticForm(ZZ, 3, [2, 1, 2, 2, 1, 3])
        sage: sorted(Q.automorphisms())                                                 # needs sage.libs.gap
        [
        [-1  0  0]  [1 0 0]
        [ 0 -1  0]  [0 1 0]
        [ 0  0 -1], [0 0 1]
        ]
    """
    return [x.matrix() for x in self.automorphism_group()]


def number_of_automorphisms(self):
    r"""
    Return the number of automorphisms (of det `1` and `-1`) of
    the quadratic form.

    OUTPUT: integer `\geq 2`

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [1, 0, 0, 1, 0, 1], unsafe_initialization=True)
        sage: Q.number_of_automorphisms()
        48

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.number_of_automorphisms()
        384
        sage: 2^4 * factorial(4)
        384
    """
    try:
        return self.__number_of_automorphisms
    except AttributeError:
        self._compute_automorphisms()
        return self.__number_of_automorphisms


def set_number_of_automorphisms(self, num_autos):
    r"""
    Set the number of automorphisms to be the value given.  No error
    checking is performed, to this may lead to erroneous results.

    The fact that this result was set externally is recorded in the
    internal list of external initializations, accessible by the
    method :meth:`list_external_initializations`.

    OUTPUT: none

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1, 1, 1])
        sage: Q.list_external_initializations()
        []
        sage: Q.set_number_of_automorphisms(-3)
        sage: Q.number_of_automorphisms()
        -3
        sage: Q.list_external_initializations()
        ['number_of_automorphisms']
    """
    self.__number_of_automorphisms = num_autos
    text = 'number_of_automorphisms'
    if text not in self._external_initialization_list:
        self._external_initialization_list.append(text)



r"""
Sparse Matrices over a general ring

EXAMPLES::

    sage: R.<x> = PolynomialRing(QQ)
    sage: M = MatrixSpace(QQ['x'], 2, 3, sparse=True); M
    Full MatrixSpace of 2 by 3 sparse matrices over
     Univariate Polynomial Ring in x over Rational Field
    sage: a = M(range(6)); a
    [0 1 2]
    [3 4 5]
    sage: b = M([x^n for n in range(6)]); b
    [  1   x x^2]
    [x^3 x^4 x^5]
    sage: a * b.transpose()
    [            2*x^2 + x           2*x^5 + x^4]
    [      5*x^2 + 4*x + 3 5*x^5 + 4*x^4 + 3*x^3]
    sage: pari(a)*pari(b.transpose())                                                   # needs sage.libs.pari
    [2*x^2 + x, 2*x^5 + x^4; 5*x^2 + 4*x + 3, 5*x^5 + 4*x^4 + 3*x^3]
    sage: c = copy(b); c
    [  1   x x^2]
    [x^3 x^4 x^5]
    sage: c[0,0] = 5; c
    [  5   x x^2]
    [x^3 x^4 x^5]
    sage: b[0,0]
    1
    sage: c.dict()
    {(0, 0): 5, (0, 1): x, (0, 2): x^2, (1, 0): x^3, (1, 1): x^4, (1, 2): x^5}
    sage: c.list()
    [5, x, x^2, x^3, x^4, x^5]
    sage: c.rows()
    [(5, x, x^2), (x^3, x^4, x^5)]
    sage: TestSuite(c).run()
    sage: d = c.change_ring(CC['x']); d
    [5.00000000000000                x              x^2]
    [             x^3              x^4              x^5]
    sage: latex(c)
    \left(\begin{array}{rrr}
    5 & x & x^{2} \\
    x^{3} & x^{4} & x^{5}
    \end{array}\right)
    sage: c.sparse_rows()
    [(5, x, x^2), (x^3, x^4, x^5)]
    sage: d = c.dense_matrix(); d
    [  5   x x^2]
    [x^3 x^4 x^5]
    sage: parent(d)
    Full MatrixSpace of 2 by 3 dense matrices
     over Univariate Polynomial Ring in x over Rational Field
    sage: c.sparse_matrix() is c
    True
    sage: c.is_sparse()
    True
"""
cimport sage.matrix.matrix_sparse as matrix_sparse
cimport sage.structure.element
from sage.matrix.args cimport MatrixArgs_init


cdef class Matrix_generic_sparse(matrix_sparse.Matrix_sparse):
    r"""
    Generic sparse matrix.

    The ``Matrix_generic_sparse`` class derives from
    :class:`~sage.matrix.matrix_sparse.Matrix_sparse`, and defines functionality
    for sparse matrices over any base ring. A generic sparse matrix is
    represented using a dictionary whose keys are pairs of integers `(i,j)` and
    values in the base ring. The values of the dictionary must never be zero.

    EXAMPLES::

        sage: R.<a,b> = PolynomialRing(ZZ,'a,b')
        sage: M = MatrixSpace(R,5,5,sparse=True)
        sage: M({(0,0):5*a+2*b, (3,4): -a})
        [5*a + 2*b         0         0         0         0]
        [        0         0         0         0         0]
        [        0         0         0         0         0]
        [        0         0         0         0        -a]
        [        0         0         0         0         0]
        sage: M(3)
        [3 0 0 0 0]
        [0 3 0 0 0]
        [0 0 3 0 0]
        [0 0 0 3 0]
        [0 0 0 0 3]
        sage: V = FreeModule(ZZ, 5,sparse=True)
        sage: m = M([V({0:3}), V({2:2, 4:-1}), V(0), V(0), V({1:2})])
        sage: m
        [ 3  0  0  0  0]
        [ 0  0  2  0 -1]
        [ 0  0  0  0  0]
        [ 0  0  0  0  0]
        [ 0  2  0  0  0]

    .. NOTE::

        The datastructure can potentially be optimized. Firstly, as noticed in
        :issue:`17663`, we lose time in using 2-tuples to store indices.
        Secondly, there is no fast way to access nonzero elements in a given
        row/column.
    """
    def __cinit__(self):
        self._entries = {}  # crucial so that pickling works
        self._zero = self._base_ring.zero()

    def __init__(self, parent, entries=None, copy=None, bint coerce=True):
        r"""
        Create a sparse matrix over the given base ring.

        INPUT:

        - ``parent`` -- a matrix space

        - ``entries`` -- see :func:`matrix`

        - ``copy`` -- ignored (for backwards compatibility)

        - ``coerce`` -- if ``False``, assume without checking that the
          entries lie in the base ring

        TESTS::

            sage: R.<a> = PolynomialRing(ZZ,'a')
            sage: M = MatrixSpace(R,2,3,sparse=True)
            sage: m = M([4,1,0,0,0,2]); m
            [4 1 0]
            [0 0 2]
            sage: m2 = copy(m)
            sage: m2[0,0] = -1
            sage: m[0,0]
            4
            sage: loads(dumps(m)) == m
            True

            sage: R2.<a,b> = PolynomialRing(QQ)
            sage: M2 = MatrixSpace(R2,2,3,sparse=True)
            sage: M2(m)
            [4 1 0]
            [0 0 2]
            sage: M2.has_coerce_map_from(M)
            True

            sage: M3 = M2.change_ring(R2.fraction_field())
            sage: M3.has_coerce_map_from(M2)
            True

        Check that it is not possible to use wrong indices::

            sage: M = MatrixSpace(R,2,2,sparse=True)
            sage: M({(3,0): 1})
            Traceback (most recent call last):
            ...
            IndexError: invalid row index 3
            sage: M({(0,-3): 1})
            Traceback (most recent call last):
            ...
            IndexError: invalid column index -3

        But negative indices are valid and wrap around::

            sage: M({(-1,-1): 1})
            [0 0]
            [0 1]
        """
        ma = MatrixArgs_init(parent, entries)
        self._entries = ma.dict(coerce)

    def __bool__(self):
        r"""
        Test whether this matrix is nonzero.

        TESTS::

            sage: R.<a,b> = Zmod(5)['a','b']
            sage: m = matrix(R,3,4,sparse=True)
            sage: bool(m)      # indirect doctest
            False
            sage: m[0,3] = 1
            sage: bool(m)      # indirect doctest
            True
            sage: m[0,3] = 0   # indirect doctest
            sage: bool(m)
            False
            sage: m.is_zero()  # indirect doctest
            True
        """
        return bool(self._entries)

    cdef set_unsafe(self, Py_ssize_t i, Py_ssize_t j, value):
        if not value:
            try:
                del self._entries[(i, j)]
            except KeyError:
                pass
        else:
            self._entries[(i, j)] = value

    cdef get_unsafe(self, Py_ssize_t i, Py_ssize_t j):
        return self._entries.get((i, j), self._zero)

    cdef copy_from_unsafe(self, Py_ssize_t iDst, Py_ssize_t jDst, src, Py_ssize_t iSrc, Py_ssize_t jSrc):
        r"""
        Copy the ``(iSrc, jSrc)`` entry of ``src`` into the ``(iDst, jDst)``
        entry of ``self``.

        INPUT:

        - ``iDst`` - the row to be copied to in ``self``.
        - ``jDst`` - the column to be copied to in ``self``.
        - ``src`` - the matrix to copy from. Should be a Matrix_generic_sparse
                    with the same base ring as ``self``.
        - ``iSrc``  - the row to be copied from in ``src``.
        - ``jSrc`` - the column to be copied from in ``src``.

        TESTS::

            sage: K.<z> = GF(9)
            sage: m = matrix(K,3,4,[((i%9)//3)*z + i%3 if is_prime(i) else 0 for i in range(12)],sparse=True)
            sage: m
            [      0       0       2       z]
            [      0   z + 2       0 2*z + 1]
            [      0       0       0       2]
            sage: m.transpose()
            [      0       0       0]
            [      0   z + 2       0]
            [      2       0       0]
            [      z 2*z + 1       2]
            sage: m.matrix_from_rows([0,2])
            [0 0 2 z]
            [0 0 0 2]
            sage: m.matrix_from_columns([1,3])
            [      0       z]
            [  z + 2 2*z + 1]
            [      0       2]
            sage: m.matrix_from_rows_and_columns([1,2],[0,3])
            [      0 2*z + 1]
            [      0       2]
        """
        cdef Matrix_generic_sparse _src = <Matrix_generic_sparse>src
        if (iSrc, jSrc) in _src._entries:
            self._entries[(iDst, jDst)] = _src._entries.get((iSrc, jSrc), _src._zero)
        elif (iDst, jDst) in self._entries:
            del self._entries[(iDst, jDst)]

    cdef bint get_is_zero_unsafe(self, Py_ssize_t i, Py_ssize_t j) except -1:
        """
        Return 1 if the entry ``(i, j)`` is zero, otherwise 0.

        EXAMPLES::

            sage: R.<a,b> = Zmod(5)['a','b']
            sage: m = matrix(R,2,4, {(1,3): a, (0,0):b}, sparse=True)
            sage: m.zero_pattern_matrix()  # indirect doctest
            [0 1 1 1]
            [1 1 1 0]
        """
        return (i, j) not in self._entries

    def _pickle(self):
        version = 0
        return self._entries, version

    def _unpickle(self, data, int version):
        """
        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: a = matrix(R, [[1,10],[3,4]],sparse=True); a
            [ 1 10]
            [ 3  4]
            sage: loads(dumps(a)) == a
            True
        """
        if version == 0:
            self._entries = data
            self._zero = self._base_ring(0)
        else:
            raise RuntimeError(f"unknown matrix version (={version})")

    ########################################################################
    # LEVEL 2 functionality
    # x  * cdef _add_
    #    * cdef _mul_
    #    * cpdef _richcmp_
    #    * __neg__
    #    * __invert__
    # x  * __copy__
    #    * _multiply_classical
    # x  * _list -- copy of the list of underlying elements
    # x  * _dict -- copy of the sparse dictionary of underlying elements
    ########################################################################

    cpdef _add_(self, _other):
        """
        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: a = matrix(R, [[1,10],[3,4]],sparse=True); a
            [ 1 10]
            [ 3  4]
            sage: a+a
            [ 2 20]
            [ 6  8]

        ::

            sage: a = matrix(R, [[1,10,-5/3],[2/8,3,4]], sparse=True); a
            [   1   10 -5/3]
            [ 1/4    3    4]
            sage: a+a
            [    2    20 -10/3]
            [  1/2     6     8]

        TESTS:

        Adding two subclasses should produce an instance of the subclass; not
        a generic matrix::

            sage: A = matrix(RDF, [[1]], sparse=True)
            sage: B = matrix(RDF, [[2]], sparse=True)
            sage: C = matrix(ZZ,  [[3]], sparse=True)
            sage: D = matrix(CDF, [[4]], sparse=True)
            sage: (A+B).__class__ == A.__class__
            True
            sage: (B+A).__class__ == A.__class__
            True
            sage: (A+C).__class__ == A.__class__
            True
            sage: (C+A).__class__ == A.__class__
            True
            sage: (A+D).__class__ == D.__class__
            True
        """
        # Compute the sum of two sparse matrices.
        # This is complicated because of how we represent sparse matrices.
        # Algorithm:
        #   1. Sort both entry coefficient lists.
        #   2. March through building a new list, adding when the two i,j are the same.
        cdef Py_ssize_t i, j, len_v, len_w
        cdef Matrix_generic_sparse other
        other = <Matrix_generic_sparse> _other
        cdef list v = sorted(self._entries.items())
        cdef list w = sorted(other._entries.items())
        s = {}
        i = 0  # pointer into self
        j = 0  # pointer into other
        len_v = len(v)
        len_w = len(w)
        while i < len_v and j < len_w:
            vi = v[i][0]
            wj = w[j][0]
            if vi < wj:
                s[vi] = v[i][1]
                i += 1
            elif vi > wj:
                s[wj] = w[j][1]
                j += 1
            else:  # equal
                sm = v[i][1] + w[j][1]
                if sm:
                    s[vi] = sm
                i += 1
                j += 1
        while i < len(v):
            s[v[i][0]] = v[i][1]
            i += 1
        while j < len(w):
            s[w[j][0]] = w[j][1]
            j += 1

        # Use the parent of the left-hand summand as the parent of the
        # result. This allows you to (for example) add two sparse RDF
        # matrices together and get another one back, rather than
        # getting a generic sparse matrix back.
        MS = self._parent
        cdef type t = <type>type(self)
        A = <Matrix_generic_sparse>t.__new__(t, MS)
        A._entries = s
        A._zero = self._zero
        return A

    def __copy__(self):
        A = self.__class__(self._parent, self._entries, copy = True, coerce=False)
        if self._subdivisions is not None:
            A.subdivide(*self.subdivisions())
        return A

    def _list(self):
        """
        Return all entries of ``self`` as a list of numbers of rows times
        number of columns entries.
        """
        cdef Py_ssize_t i, j
        cdef list v = self.fetch('list')
        if v is None:
            v = [self._zero]*(self._nrows * self._ncols)
            for (i, j), x in self._entries.items():
                v[i*self._ncols + j] = x
            self.cache('list', v)
        return v

    def _dict(self):
        """
        Return the underlying dictionary of ``self``.

        This is used in comparisons.

        TESTS::

            sage: R.<a,b> = Zmod(6)[]
            sage: M = MatrixSpace(R, 3, 4)
            sage: m = M({(0,3): a+3*b*a, (1,1): -b})
            sage: m == m    # indirect doctest
            True
            sage: M(0) == m # indirect doctest
            False
        """
        return self._entries

    ########################################################################
    # LEVEL 3 functionality -- matrix windows, etc.
    ########################################################################

    def _nonzero_positions_by_row(self, copy=True):
        r"""
        TESTS::

            sage: R.<a> = PolynomialRing(Zmod(8), 'a')
            sage: M = MatrixSpace(R,4,3,sparse=True)
            sage: m = M({(3,0): 1, (3,1): 2*a^2 + 1, (2,0): -1, (0,1): -2})
            sage: m._nonzero_positions_by_row()
            [(0, 1), (2, 0), (3, 0), (3, 1)]
        """
        cdef list v = self.fetch('nonzero_positions')
        if v is None:
            v = sorted(self._entries)
            self.cache('nonzero_positions', v)
        if copy:
            return v[:]
        return v

    def _nonzero_positions_by_column(self, copy=True):
        r"""
        TESTS::

            sage: R.<a> = PolynomialRing(Zmod(8), 'a')
            sage: M = MatrixSpace(R,4,3,sparse=True)
            sage: m = M({(3,0): 1, (3,1): 2*a^2 + 1, (2,0): -1, (0,1): -2})
            sage: m._nonzero_positions_by_column()
            [(2, 0), (3, 0), (0, 1), (3, 1)]
        """
        cdef list v = self.fetch('nonzero_positions_by_column')
        if v is None:
            v = sorted(self._entries, key=lambda x: (x[1], x[0]))
            self.cache('nonzero_positions_by_column', v)
        if copy:
            return v[:]
        return v


####################################################################################
# Various helper functions
####################################################################################

def Matrix_sparse_from_rows(X):
    """
    INPUT:

    - ``X`` -- nonempty list of ``SparseVector`` rows

    OUTPUT: ``Sparse_matrix`` with those rows

    EXAMPLES::

        sage: V = VectorSpace(QQ,20,sparse=True)
        sage: v = V(0)
        sage: v[9] = 4
        sage: from sage.matrix.matrix_generic_sparse import Matrix_sparse_from_rows
        sage: Matrix_sparse_from_rows([v])
        [0 0 0 0 0 0 0 0 0 4 0 0 0 0 0 0 0 0 0 0]
        sage: Matrix_sparse_from_rows([v, v, v, V(0)])
        [0 0 0 0 0 0 0 0 0 4 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 4 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 4 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
    """
    cdef Py_ssize_t i, j

    if not isinstance(X, (list, tuple)):
        raise TypeError("X (=%s) must be a list or tuple" % X)
    if not X:
        raise ArithmeticError("X must be nonempty")

    from sage.matrix import matrix_space
    entries = {}
    R = X[0].base_ring()
    ncols = X[0].degree()
    for i from 0 <= i < len(X):
        for j, x in X[i].items():
            entries[(i, j)] = x
    M = matrix_space.MatrixSpace(R, len(X), ncols, sparse=True)
    return M(entries, coerce=False, copy=False)
